# coding: utf-8
# cython: boundscheck=False
# cython: nonecheck=False
# cython: cdivision=True
# cython: wraparound=False
# cython: profile=False

from __future__ import division, print_function

__author__ = "adrn <adrn@astro.columbia.edu>"

# Standard library
from collections import OrderedDict

# Third-party
from astropy.extern import six
from astropy.utils import InheritDocstrings
import numpy as np
cimport numpy as np
np.import_array()
import cython
cimport cython

from libc.stdio cimport printf

# Project
from .core import PotentialBase, CompositePotential
from ..util import atleast_2d
from ..units import DimensionlessUnitSystem

cdef extern from "math.h":
    double sqrt(double x) nogil
    double fabs(double x) nogil

cdef extern from "src/cpotential.h":
    enum:
        MAX_N_COMPONENTS = 16

    ctypedef double (*densityfunc)(double t, double *pars, double *q) nogil
    ctypedef double (*valuefunc)(double t, double *pars, double *q) nogil
    ctypedef void (*gradientfunc)(double t, double *pars, double *q, double *grad) nogil

    ctypedef struct CPotential:
        int n_components
        int n_dim
        densityfunc density[MAX_N_COMPONENTS]
        valuefunc value[MAX_N_COMPONENTS]
        gradientfunc gradient[MAX_N_COMPONENTS]
        int n_params[MAX_N_COMPONENTS]
        double *parameters[MAX_N_COMPONENTS]

    double c_value(CPotential *p, double t, double *q) nogil
    double c_density(CPotential *p, double t, double *q) nogil
    void c_gradient(CPotential *p, double t, double *q, double *grad) nogil

    double c_d_dr(CPotential *p, double t, double *q, double *epsilon) nogil
    double c_d2_dr2(CPotential *p, double t, double *q, double *epsilon) nogil
    double c_mass_enclosed(CPotential *p, double t, double *q, double G, double *epsilon) nogil

__all__ = ['CPotentialBase']

cdef class CPotentialWrapper:
    """
    Wrapper class for C implementation of potentials. At the C layer, potentials
    are effectively struct's that maintain pointers to functions specific to a
    given potential. This provides a Cython wrapper around this C implementation.
    """

    cpdef value(self, double[:,::1] q, double t=0.):
        """
        CAUTION: Interpretation of axes is different here! We need the
        arrays to be C ordered and easy to iterate over, so here the
        axes are (norbits, ndim).
        """
        cdef int norbits, ndim, i

        if q.ndim != 2:
            raise ValueError("Coordinate array q must have 2 dimensions")

        norbits = q.shape[0]
        ndim = q.shape[1]

        cdef double [::1] pot = np.zeros((norbits,))
        for i in range(norbits):
            pot[i] = c_value(&(self.cpotential), t, &q[i,0])

        return np.array(pot)

    cpdef density(self, double[:,::1] q, double t=0.):
        """
        CAUTION: Interpretation of axes is different here! We need the
        arrays to be C ordered and easy to iterate over, so here the
        axes are (norbits, ndim).
        """
        cdef int norbits, ndim, i

        if q.ndim != 2:
            raise ValueError("Coordinate array q must have 2 dimensions")

        norbits = q.shape[0]
        ndim = q.shape[1]

        cdef double [::1] dens = np.zeros((norbits,))
        for i in range(norbits):
            dens[i] = c_density(&(self.cpotential), t, &q[i,0])

        return np.array(dens)

    cpdef gradient(self, double[:,::1] q, double t=0.):
        """
        CAUTION: Interpretation of axes is different here! We need the
        arrays to be C ordered and easy to iterate over, so here the
        axes are (norbits, ndim).
        """
        cdef int norbits, ndim, i

        if q.ndim != 2:
            raise ValueError("Coordinate array q must have 2 dimensions")

        norbits = q.shape[0]
        ndim = q.shape[1]

        cdef double [:,::1] grad = np.zeros((norbits, ndim))
        for i in range(norbits):
            c_gradient(&(self.cpotential), t, &q[i,0], &grad[i,0])

        return np.array(grad)

    # cpdef hessian(self, double[:,::1] q, double t=0.):
    #     """
    #     CAUTION: Interpretation of axes is different here! We need the
    #     arrays to be C ordered and easy to iterate over, so here the
    #     axes are (norbits, ndim).
    #     """
    #     cdef int norbits, ndim, i

    #     if q.ndim != 2:
    #         raise ValueError("Coordinate array q must have 2 dimensions")

    #     norbits = q.shape[0]
    #     ndim = q.shape[1]

    #     cdef double [::1] hess = np.zeros(q.shape + (ndim,)))
    #     for i in range(norbits):
    #         c_hessian(&(self.cpotential), t, &q[i,0], &hess[i,0,0])

    #     return np.array(hess)

    # Second order functionality

    cpdef d_dr(self, double[:,::1] q, double G, double t=0.):
        """
        CAUTION: Interpretation of axes is different here! We need the
        arrays to be C ordered and easy to iterate over, so here the
        axes are (norbits, ndim).
        """
        cdef:
            int i
            int norbits = q.shape[0]
            double [::1] epsilon = np.zeros(3, dtype=np.float64)
            double [::1] dr = np.zeros(norbits, dtype=np.float64)

        for i in range(norbits):
            dr[i] = c_d_dr(&(self.cpotential), t, &q[i,0], &epsilon[0])

        return np.array(dr)

    cpdef d2_dr2(self, double[:,::1] q, double G, double t=0.):
        """
        CAUTION: Interpretation of axes is different here! We need the
        arrays to be C ordered and easy to iterate over, so here the
        axes are (norbits, ndim).
        """
        cdef:
            int i
            int norbits = q.shape[0]
            double [::1] epsilon = np.zeros(3, dtype=np.float64)
            double [::1] dr2 = np.zeros(norbits, dtype=np.float64)

        for i in range(norbits):
            dr2[i] = c_d2_dr2(&(self.cpotential), t, &q[i,0], &epsilon[0])

        return np.array(dr2)

    cpdef mass_enclosed(self, double[:,::1] q, double G, double t=0.):
        """
        CAUTION: Interpretation of axes is different here! We need the
        arrays to be C ordered and easy to iterate over, so here the
        axes are (norbits, ndim).
        """
        cdef:
            int i
            int norbits = q.shape[0]
            double [::1] epsilon = np.zeros(3, dtype=np.float64)
            double [::1] mass = np.zeros(norbits, dtype=np.float64)

        for i in range(norbits):
            mass[i] = c_mass_enclosed(&(self.cpotential), t, &q[i,0], G, &epsilon[0])

        return np.array(mass)

    # For pickling in Python 2
    def __reduce__(self):
        return (self.__class__, (self._params[0], list(self._params[1:])))

class CPotentialBase(PotentialBase):
    """
    A baseclass for defining gravitational potentials implemented in C.
    """

    def __init__(self, parameters, units, Wrapper=None):
        super(CPotentialBase, self).__init__(parameters, units=units)

        c_params = []
        for k,v in self.parameters.items():
            c_params.append(self.parameters[k].value)
        self.c_parameters = np.array(c_params)

        if Wrapper is None:
            # magic to set the c_instance attribute based on the name of the class
            wrapper_name = '{}Wrapper'.format(self.__class__.__name__.replace('Potential', ''))

            from .builtin import cybuiltin
            Wrapper = getattr(cybuiltin, wrapper_name)

        self.c_instance = Wrapper(self.G, self.c_parameters)


    def _value(self, q, t=0.):
        sh = q.shape
        q = np.ascontiguousarray(q.reshape(sh[0],np.prod(sh[1:])).T)
        return self.c_instance.value(q, t=t).reshape(sh[1:])

    def _density(self, q, t=0.):
        sh = q.shape
        q = np.ascontiguousarray(q.reshape(sh[0],np.prod(sh[1:])).T)
        try:
            return self.c_instance.density(q, t=t).reshape(sh[1:])
        except AttributeError,TypeError:
            # TODO: if no density function, should this numerically esimate
            #   the density?
            raise ValueError("Potential C instance has no defined "
                             "density function")

    def _gradient(self, q, t=0.):
        sh = q.shape
        q = np.ascontiguousarray(q.reshape(sh[0],np.prod(sh[1:])).T)
        try:
            return self.c_instance.gradient(q, t=t).T.reshape(sh)
        except AttributeError,TypeError:
            raise ValueError("Potential C instance has no defined "
                             "gradient function")

    def _hessian(self, q, t=0.):
        sh = q.shape
        q = np.ascontiguousarray(q.reshape(sh[0],np.prod(sh[1:])).T)
        try:
            return self.c_instance.hessian(q, t=t) # TODO: return shape?
        except AttributeError,TypeError:
            raise ValueError("Potential C instance has no defined "
                             "Hessian function")

    # ----------------------------------------------------------
    # Overwrite the Python potential method to use Cython method
    def mass_enclosed(self, q, t=0.):
        """
        mass_enclosed(q, t)

        Estimate the mass enclosed within the given position by assuming the potential
        is spherical. This is not so good!

        Parameters
        ----------
        q : array_like, numeric
            Position to compute the mass enclosed.
        """
        if isinstance(self.units, DimensionlessUnitSystem):
            raise ValueError("No units specified when creating potential object.")

        q = atleast_2d(q, insert_axis=1)
        sh = q.shape
        q = np.ascontiguousarray(q.reshape(sh[0],np.prod(sh[1:])).T)
        try:
            menc = self.c_instance.mass_enclosed(q, self.G, t=t)
        except AttributeError,TypeError:
            raise ValueError("Potential C instance has no defined "
                             "mass_enclosed function")
        return menc.reshape(sh[1:]) * self.units['mass']



# coding: utf-8
"""
    Test the core observational utilities.
"""

from __future__ import absolute_import, unicode_literals, division, print_function

__author__ = "adrn <adrn@astro.columbia.edu>"

# Standard library
import os, sys

# Third-party
import astropy.units as u
import numpy as np
import pytest

from ..core import *




# coding: utf-8

""" Test ...  """

from __future__ import division, print_function

__author__ = "adrn <adrn@astro.columbia.edu>"

# Third-party
from astropy.coordinates import SphericalRepresentation, Galactic
import astropy.units as u
import numpy as np
import pytest
import scipy.optimize as so

# Project
from ..core import CartesianPhaseSpacePosition
from ..orbit import *
from ...integrate import DOPRI853Integrator
from ...potential import HernquistPotential, LogarithmicPotential, KeplerPotential
from ...units import galactic, solarsystem

def make_known_orbit(tmpdir, x, vx, potential, name):
    # See Binney & Tremaine (2008) Figure 3.8 and 3.9
    E = -0.337
    y = 0.
    vy = np.sqrt(2*(E - potential.value([x,y,0.]).value))[0]

    w = [x,y,0.,vx,vy,0.]
    orbit = potential.integrate_orbit(w, dt=0.05, n_steps=10000)

    # fig,ax = pl.subplots(1,1)
    # ax.plot(ws[0], ws[1])
    # fig = plot_orbits(ws, linestyle='none', alpha=0.1)
    # fig.savefig(os.path.join(str(tmpdir), "{}.png".format(name)))
    # logger.debug(os.path.join(str(tmpdir), "{}.png".format(name)))

    return orbit

def test_circulation(tmpdir):

    potential = LogarithmicPotential(v_c=1., r_h=0.14, q1=1., q2=0.9, q3=1.,
                                     units=galactic)

    # individual
    w1 = make_known_orbit(tmpdir, 0.5, 0., potential, "loop")
    circ = w1.circulation()
    assert circ.shape == (3,)
    assert circ.sum() == 1

    w2 = make_known_orbit(tmpdir, 0., 1.5, potential, "box")
    circ = w2.circulation()
    assert circ.shape == (3,)
    assert circ.sum() == 0

    # try also for both, together
    w3 = combine((w1,w2))
    circ = w3.circulation()
    assert circ.shape == (3,2)
    assert np.allclose(circ.sum(axis=0), [1,0])

def test_align_circulation():

    t = np.linspace(0,100,1024)
    w = np.zeros((6,1024,4))

    # loop around x axis
    w[1,:,0] = np.cos(t)
    w[2,:,0] = np.sin(t)
    w[4,:,0] = -np.sin(t)
    w[5,:,0] = np.cos(t)

    # loop around y axis
    w[0,:,1] = -np.cos(t)
    w[2,:,1] = np.sin(t)
    w[3,:,1] = np.sin(t)
    w[5,:,1] = np.cos(t)

    # loop around z axis
    w[0,:,2] = np.cos(t)
    w[1,:,2] = np.sin(t)
    w[3,:,2] = -np.sin(t)
    w[4,:,2] = np.cos(t)

    # box
    w[0,:,3] = np.cos(t)
    w[1,:,3] = -np.cos(0.5*t)
    w[2,:,3] = np.cos(0.25*t)
    w[3,:,3] = -np.sin(t)
    w[4,:,3] = 0.5*np.sin(0.5*t)
    w[5,:,3] = -0.25*np.sin(0.25*t)

    # First, individually
    for i in range(w.shape[2]):
        orb = CartesianOrbit.from_w(w[...,i], units=galactic)
        new_orb = orb.align_circulation_with_z()
        circ = new_orb.circulation()

        if i == 3:
            assert np.sum(circ) == 0
        else:
            assert circ[2] == 1.

    # all together now
    orb = CartesianOrbit.from_w(w, units=galactic)
    circ = orb.circulation()
    assert circ.shape == (3,4)

    new_orb = orb.align_circulation_with_z()
    new_circ = new_orb.circulation()
    assert np.all(new_circ[2,:3] == 1.)
    assert np.all(new_circ[:,3] == 0.)

# Tests below should be fixed a bit...
def test_initialize():

    with pytest.raises(ValueError):
        x = np.random.random(size=(3,10))
        v = np.random.random(size=(3,8))
        CartesianOrbit(pos=x, vel=v)

    with pytest.raises(ValueError):
        x = np.random.random(size=(3,10))
        v = np.random.random(size=(3,10))
        t = np.arange(8)
        CartesianOrbit(pos=x, vel=v, t=t)

    x = np.random.random(size=(3,10))
    v = np.random.random(size=(3,10))
    o = CartesianOrbit(pos=x, vel=v)
    assert o.ndim == 3

    x = np.random.random(size=(3,10))*u.kpc
    v = np.random.random(size=(3,10))*u.km/u.s
    o = CartesianOrbit(pos=x, vel=v)
    assert o.pos.unit == u.kpc
    assert o.vel.unit == u.km/u.s

    x = np.random.random(size=(2,10))
    v = np.random.random(size=(2,10))
    o = CartesianOrbit(pos=x, vel=v)
    assert o.ndim == 2

def test_from_w():

    w = np.random.random(size=(6,10))
    o = CartesianOrbit.from_w(w, galactic)
    assert o.pos.unit == u.kpc
    assert o.vel.unit == u.kpc/u.Myr

def test_slice():

    # simple
    x = np.random.random(size=(3,10))
    v = np.random.random(size=(3,10))
    o = CartesianOrbit(pos=x, vel=v)
    new_o = o[:5]
    assert new_o.pos.shape == (3,5)

    # 1d slice on 3d
    x = np.random.random(size=(3,100,8))
    v = np.random.random(size=(3,100,8))
    t = np.arange(x.shape[1])
    o = CartesianOrbit(pos=x, vel=v, t=t)
    new_o = o[:5]
    assert new_o.pos.shape == (3,5,8)
    assert new_o.t.shape == (5,)

    # 3d slice on 3d
    o = CartesianOrbit(pos=x, vel=v, t=t)
    new_o = o[:5,:4]
    assert new_o.pos.shape == (3,5,4)
    assert new_o.t.shape == (5,)

    # boolean array
    x = np.random.random(size=(3,10))
    v = np.random.random(size=(3,10))
    t = np.arange(x.shape[1])
    o = CartesianOrbit(pos=x, vel=v, t=t)
    ix = np.array([0,0,0,0,0,1,1,1,1,1]).astype(bool)
    new_o = o[ix]
    assert new_o.shape == (sum(ix),)
    assert new_o.t.shape == (5,)

    # boolean array - 3D
    x = np.random.random(size=(3,10,4))
    v = np.random.random(size=(3,10,4))
    t = np.arange(x.shape[1])
    o = CartesianOrbit(pos=x, vel=v, t=t)
    ix = np.array([0,0,0,0,0,1,1,1,1,1]).astype(bool)
    new_o = o[ix]
    assert new_o.shape == (sum(ix),x.shape[-1])
    assert new_o.t.shape == (5,)

    # integer array
    x = np.random.random(size=(3,10))
    v = np.random.random(size=(3,10))
    t = np.arange(x.shape[1])
    o = CartesianOrbit(pos=x, vel=v, t=t)
    ix = np.array([0,3,5])
    new_o = o[ix]
    assert new_o.shape == (len(ix),)
    assert new_o.t.shape == (len(ix),)

def test_represent_as():

    # simple / unitless
    x = np.random.random(size=(3,10))
    v = np.random.random(size=(3,10))
    o = CartesianOrbit(pos=x, vel=v)
    sph_pos, sph_vel = o.represent_as(SphericalRepresentation)

    assert sph_pos.distance.unit == u.dimensionless_unscaled
    assert sph_vel.unit == u.dimensionless_unscaled

    # simple / with units
    x = np.random.random(size=(3,10))*u.kpc
    v = np.random.normal(0.,100.,size=(3,10))*u.km/u.s
    o = CartesianOrbit(pos=x, vel=v)
    sph_pos, sph_vel = o.represent_as(SphericalRepresentation)
    assert sph_pos.distance.unit == u.kpc

def test_to_frame():
    # simple / unitless
    x = np.random.random(size=(3,10))
    v = np.random.random(size=(3,10))
    o = CartesianOrbit(pos=x, vel=v)

    with pytest.raises(u.UnitConversionError):
        o.to_frame(Galactic)

    # simple / with units
    x = np.random.random(size=(3,10))*u.kpc
    v = np.random.normal(0.,100.,size=(3,10))*u.km/u.s
    o = CartesianOrbit(pos=x, vel=v)
    coo,vel = o.to_frame(Galactic)
    assert coo.name == 'galactic'

    # simple / with units and time
    x = np.random.random(size=(3,128,10))*u.kpc
    v = np.random.normal(0.,100.,size=(3,128,10))*u.km/u.s
    o = CartesianOrbit(pos=x, vel=v)
    coo,vel = o.to_frame(Galactic)
    assert coo.name == 'galactic'

def test_w():
    # simple / unitless
    x = np.random.random(size=(3,10))
    v = np.random.random(size=(3,10))
    o = CartesianOrbit(pos=x, vel=v)
    w = o.w()
    assert w.shape == (6,10,1)

    # simple / with units
    x = np.random.random(size=(3,10))*u.kpc
    v = np.random.normal(0.,100.,size=(3,10))*u.km/u.s
    o = CartesianOrbit(pos=x, vel=v)
    with pytest.raises(ValueError):
        o.w()
    w = o.w(units=galactic)
    assert np.allclose(x.value, w[:3,:,0])
    assert np.allclose(v.value, (w[3:,:,0]*u.kpc/u.Myr).to(u.km/u.s).value)

    # simple / with units and potential
    p = HernquistPotential(units=galactic, m=1E11, c=0.25)
    x = np.random.random(size=(3,10))*u.kpc
    v = np.random.normal(0.,100.,size=(3,10))*u.km/u.s
    o = CartesianOrbit(pos=x, vel=v, potential=p)
    w = o.w()
    assert np.allclose(x.value, w[:3,:,0])
    assert np.allclose(v.value, (w[3:,:,0]*u.kpc/u.Myr).to(u.km/u.s).value)

    w = o.w(units=solarsystem)
    assert np.allclose(x.value, (w[:3,:,0]*u.au).to(u.kpc).value)
    assert np.allclose(v.value, (w[3:,:,0]*u.au/u.yr).to(u.km/u.s).value)

def test_energy():
    # with units
    x = np.random.random(size=(3,10))*u.kpc
    v = np.random.normal(0.,100.,size=(3,10))*u.km/u.s
    o = CartesianOrbit(pos=x, vel=v)
    KE = o.kinetic_energy()
    assert KE.unit == (o.vel.unit)**2
    assert KE.shape == o.pos.shape[1:]

    # with units and potential
    p = HernquistPotential(units=galactic, m=1E11, c=0.25)
    x = np.random.random(size=(3,10))*u.kpc
    v = np.random.normal(0.,100.,size=(3,10))*u.km/u.s
    o = CartesianOrbit(pos=x, vel=v, potential=p)
    PE = o.potential_energy()
    E = o.energy()

def test_angular_momentum():
    # with units
    x = np.random.random(size=(3,10))*u.kpc
    v = np.random.normal(0.,100.,size=(3,10))*u.km/u.s
    o = CartesianOrbit(pos=x, vel=v)
    L = o.angular_momentum()
    assert L.unit == (o.vel.unit*o.pos.unit)
    assert L.shape == o.pos.shape

def test_eccentricity():
    pot = KeplerPotential(m=1., units=solarsystem)
    w0 = CartesianPhaseSpacePosition(pos=[1,0,0.]*u.au,
                                     vel=[0.,2*np.pi,0.]*u.au/u.yr)
    w = pot.integrate_orbit(w0, dt=0.01, n_steps=10000, Integrator=DOPRI853Integrator)
    e = w.eccentricity()
    assert np.abs(e) < 1E-3

def test_apocenter_pericenter():
    pot = KeplerPotential(m=1., units=solarsystem)
    w0 = CartesianPhaseSpacePosition(pos=[1,0,0.]*u.au,
                                     vel=[0.,1.5*np.pi,0.]*u.au/u.yr)

    w = pot.integrate_orbit(w0, dt=0.01, n_steps=10000, Integrator=DOPRI853Integrator)
    apo = w.apocenter()
    per = w.pericenter()

    assert apo.unit == u.au
    assert per.unit == u.au
    assert apo > per

    # see if they're where we expect
    E = np.mean(w.energy()).decompose(pot.units).value
    # Phi = np.mean(w.potential_energy()).value
    L = np.mean(np.sqrt(np.sum(w.angular_momentum()**2, axis=0))).decompose(pot.units).value
    def func(r):
        val = 2*(E-pot.value([r,0,0]).value[0]) - L**2/r**2
        return val

    pred_apo = so.brentq(func, 0.9, 1.0)
    pred_per = so.brentq(func, 0.3, 0.5)

    assert np.allclose(apo.value, pred_apo, rtol=1E-2)
    assert np.allclose(per.value, pred_per, rtol=1E-2)

def test_estimate_period():
    ntimes = 16384
    for true_T_R in [1., 2., 4.123]:
        t = np.linspace(0,10.,ntimes)
        R = 0.25*np.sin(2*np.pi/true_T_R * t) + 1.
        phi = (2*np.pi * t) % (2*np.pi)

        pos = np.zeros((3,ntimes))
        pos[0] = R*np.cos(phi)
        pos[1] = R*np.sin(phi)
        vel = np.zeros_like(pos)

        orb = CartesianOrbit(pos*u.kpc, vel*u.kpc/u.Myr, t=t*u.Gyr)
        T = orb.estimate_period()
        assert np.allclose(T.value, true_T_R, rtol=1E-3)

def test_combine():

    o1 = CartesianOrbit.from_w(np.random.random(size=6), units=galactic)
    o2 = CartesianOrbit.from_w(np.random.random(size=6), units=galactic)
    o = combine((o1, o2))
    assert o.pos.shape == (3,1,2)
    assert o.vel.shape == (3,1,2)

    o1 = CartesianOrbit.from_w(np.random.random(size=(6,11,1)), units=galactic)
    o2 = CartesianOrbit.from_w(np.random.random(size=(6,11,10)), units=galactic)
    o = combine((o1, o2))
    assert o.pos.shape == (3,11,11)
    assert o.vel.shape == (3,11,11)

    o1 = CartesianOrbit.from_w(np.random.random(size=(6,20,1)), units=galactic)
    o2 = CartesianOrbit.from_w(np.random.random(size=(6,10,1)), units=galactic)
    o = combine((o1, o2), along_time_axis=True)
    assert o.pos.shape == (3,30,1)
    assert o.vel.shape == (3,30,1)

    with pytest.raises(ValueError):
        o1 = CartesianOrbit.from_w(np.random.random(size=(6,10,2)), units=galactic)
        o2 = CartesianOrbit.from_w(np.random.random(size=(6,10,1)), units=galactic)
        o = combine((o1, o2), along_time_axis=True)

    # With time
    t1 = np.linspace(0.,1,20)
    t2 = np.linspace(0.,1,10)
    o1 = CartesianOrbit.from_w(np.random.random(size=(6,20,1)), units=galactic, t=t1*u.Myr)
    o2 = CartesianOrbit.from_w(np.random.random(size=(6,10,1)), units=galactic, t=t2*u.Myr)

    with pytest.raises(ValueError):
        o = combine((o1, o2))

    o = combine((o1, o2), along_time_axis=True)
    assert o.pos.shape == (3,30,1)
    assert o.vel.shape == (3,30,1)
    assert o.t.shape == (30,)



# coding: utf-8

from __future__ import division, print_function

__author__ = "adrn <adrn@astro.columbia.edu>"

# Third-party
from astropy import log as logger
import astropy.coordinates as coord
import astropy.units as u
uno = u.dimensionless_unscaled
import numpy as np

# Project
from .plot import three_panel
from ..coordinates import velocity_transforms as vtrans
from ..coordinates import vgal_to_hel
from ..units import UnitSystem, DimensionlessUnitSystem
from ..util import atleast_2d

__all__ = ['CartesianPhaseSpacePosition', 'combine']

class PhaseSpacePosition(object):

    # ------------------------------------------------------------------------
    # Display
    # ------------------------------------------------------------------------
    def __repr__(self):
        return "<{} N={}, shape={}>".format(self.__class__.__name__, self.ndim, self.shape)

    def __str__(self):
        # TODO: should show some arrays?
        return "{} {}D, {}".format(self.__class__.__name__, self.ndim, self.shape)

    # ------------------------------------------------------------------------
    # Shape and size
    # ------------------------------------------------------------------------
    @property
    def ndim(self):
        """
        Number of coordinate dimensions. 1/2 of the phase-space dimensionality.

        .. warning::

            This is *not* the number of axes in the position or velocity
            arrays. That is accessed by doing ``{}.pos.ndim``.

        Returns
        -------
        n : int

        """.format(self.__class__.__name__)
        return self.pos.shape[0]

    @property
    def shape(self):
        """
        .. warning::

            This is *not* the shape of the position or velocity
            arrays. That is accessed by doing ``{}.pos.shape``.

        Returns
        -------
        shp : tuple

        """.format(self.__class__.__name__)
        return self.pos.shape[1:]

class CartesianPhaseSpacePosition(PhaseSpacePosition):
    """
    Represents phase-space positions in Cartesian coordinates, e.g.,
    positions and conjugate momenta (velocities).

    .. warning::

        This is an experimental class. The API may change in a future release!

    The position and velocity quantities (arrays) can have an arbitrary
    number of dimensions, but the first axis (0, 1) has special meaning::

        - `axis=0` is the coordinate dimension (e.g., x, y, z)

    So if the input position array, `pos`, has shape `pos.shape = (3, 100)`,
    this would represent 100 3D positions (`pos[0]` is `x`, `pos[1]` is `y`,
    etc.). The same is true for velocity. The position and velocity arrays
    must have the same shape.

    If the input position and velocity are arrays rather than
    :class:`~astropy.units.Quantity` objects, they are internally stored with
    dimensionles units.

    Parameters
    ----------
    pos : :class:`~astropy.units.Quantity`, array_like
        Positions. If a numpy array (e.g., has no units), this will be
        stored as a dimensionless :class:`~astropy.units.Quantity`. See
        the note above about the assumed meaning of the axes of this object.
    vel : :class:`~astropy.units.Quantity`, array_like
        Velocities. If a numpy array (e.g., has no units), this will be
        stored as a dimensionless :class:`~astropy.units.Quantity`. See
        the note above about the assumed meaning of the axes of this object.

    """
    def __init__(self, pos, vel):

        # make sure position and velocity input are 2D
        pos = atleast_2d(pos, insert_axis=1)
        vel = atleast_2d(vel, insert_axis=1)

        # make sure position and velocity have at least a dimensionless unit!
        if not hasattr(pos, 'unit'):
            pos = pos * uno

        if not hasattr(vel, 'unit'):
            vel = vel * uno

        if (pos.unit == uno and vel.unit != uno):
            logger.warning("Position unit is dimensionless but velocity unit is not."
                           "Are you sure that's what you want?")
        elif (vel.unit == uno and pos.unit != uno):
            logger.warning("Velocity unit is dimensionless but position unit is not."
                           "Are you sure that's what you want?")

        # make sure shape is the same
        for i in range(pos.ndim):
            if pos.shape[i] != vel.shape[i]:
                raise ValueError("Position and velocity must have the same shape "
                                 "{} vs {}".format(pos.shape, vel.shape))

        self.pos = pos
        self.vel = vel

    def __getitem__(self, slyce):
        if isinstance(slyce, np.ndarray) or isinstance(slyce, list):
            _slyce = np.array(slyce)
            _slyce = (slice(None),) + (slyce,)
        else:
            try:
                _slyce = (slice(None),) + tuple(slyce)
            except TypeError:
                _slyce = (slice(None),) + (slyce,)

        return self.__class__(pos=self.pos[_slyce], vel=self.vel[_slyce])

    # ------------------------------------------------------------------------
    # Convert from Cartesian to other representations
    # ------------------------------------------------------------------------
    def represent_as(self, Representation):
        """
        Represent the position and velocity of the orbit in an alternate
        coordinate system. Supports any of the Astropy coordinates
        representation classes.

        Parameters
        ----------
        Representation : :class:`~astropy.coordinates.BaseRepresentation`
            The class for the desired representation.

        Returns
        -------
        pos : `~astropy.coordinates.BaseRepresentation`
        vel : `~astropy.units.Quantity`
            The velocity in the new representation. All components
            have units of velocity -- e.g., to get an angular velocity
            in spherical representations, you'll need to divide by the radius.
        """
        if self.ndim != 3:
            raise ValueError("Representation changes require a 3D (ndim=3) "
                             "position and velocity.")

        # get the name of the desired representation
        rep_name = Representation.get_name()

        # transform the position
        car_pos = coord.CartesianRepresentation(self.pos)
        new_pos = car_pos.represent_as(Representation)

        # transform the velocity
        vfunc = getattr(vtrans, "cartesian_to_{}".format(rep_name))
        new_vel = vfunc(car_pos, self.vel)

        return new_pos, new_vel

    def to_frame(self, frame, galactocentric_frame=coord.Galactocentric(),
                 vcirc=None, vlsr=None):
        """
        Transform the orbit from Galactocentric, cartesian coordinates to
        Heliocentric coordinates in the specified Astropy coordinate frame.

        Parameters
        ----------
        frame : :class:`~astropy.coordinates.BaseCoordinateFrame`
        galactocentric_frame : :class:`~astropy.coordinates.Galactocentric`
        vcirc : :class:`~astropy.units.Quantity`
            Circular velocity of the Sun. Passed to velocity transformation.
        vlsr : :class:`~astropy.units.Quantity`
            Velocity of the Sun relative to the LSR. Passed to
            velocity transformation.

        Returns
        -------
        c : :class:`~astropy.coordinates.BaseCoordinateFrame`
            An instantiated coordinate frame.
        v : tuple
            A tuple of velocities represented as
            :class:`~astropy.units.Quantity` objects.

        """

        if self.ndim != 3:
            raise ValueError("Frame transformations require a 3D (ndim=3) "
                             "position and velocity.")

        car_pos = coord.CartesianRepresentation(self.pos)
        gc_c = galactocentric_frame.realize_frame(car_pos)
        c = gc_c.transform_to(frame)

        kw = dict()
        kw['galactocentric_frame'] = galactocentric_frame
        if vcirc is not None:
            kw['vcirc'] = vcirc
        if vlsr is not None:
            kw['vlsr'] = vlsr

        v = vgal_to_hel(c, self.vel, **kw)
        return c, v

    # Pseudo-backwards compatibility
    def w(self, units=None):
        """
        This returns a single array containing the phase-space positions.

        Parameters
        ----------
        units : `~gala.units.UnitSystem` (optional)
            The unit system to represent the position and velocity in
            before combining into the full array.

        Returns
        -------
        w : `~numpy.ndarray`
            A numpy array of all positions and velocities, without units.
            Will have shape ``(2*ndim,...)``.

        """
        if (units is None or isinstance(units, DimensionlessUnitSystem)) \
            and (self.pos.unit == uno and self.vel.unit == uno):
            units = DimensionlessUnitSystem()

        elif units is None:
            raise ValueError("A UnitSystem must be provided.")

        x = self.pos.decompose(units).value
        v = self.vel.decompose(units).value

        return np.vstack((x,v))

    @classmethod
    def from_w(cls, w, units=None, **kwargs):
        """
        Create a {name} object from a single array specifying positions
        and velocities. This is mainly for backwards-compatibility and
        it is not recommended for new users.

        Parameters
        ----------
        w : array_like
            The array of phase-space positions.
        units : `~gala.units.UnitSystem` (optional)
            The unit system that the input position+velocity array, ``w``,
            is represented in.
        **kwargs
            Any aditional keyword arguments passed to the class initializer.

        Returns
        -------
        obj : `~gala.dynamics.{name}`

        """.format(name=cls.__name__)

        ndim = w.shape[0]//2
        pos = w[:ndim]
        vel = w[ndim:]

        # TODO: this is bad form - UnitSystem should know what to do with a Dimensionless
        if units is not None and not isinstance(units, DimensionlessUnitSystem):
            units = UnitSystem(units)
            pos = pos*units['length']
            vel = vel*units['length']/units['time'] # velocity in w is from _core_units

        return cls(pos=pos, vel=vel, **kwargs)

    # ------------------------------------------------------------------------
    # Computed dynamical quantities
    # ------------------------------------------------------------------------
    def kinetic_energy(self):
        r"""
        The kinetic energy *per unit mass*:

        .. math::

            E_K = \frac{1}{2} \, |\boldsymbol{v}|^2

        Returns
        -------
        E : :class:`~astropy.units.Quantity`
            The kinetic energy.
        """
        return 0.5*np.sum(self.vel**2, axis=0)

    def potential_energy(self, potential):
        r"""
        The potential energy *per unit mass*:

        .. math::

            E_\Phi = \Phi(\boldsymbol{q})

        Parameters
        ----------
        potential : `gala.potential.PotentialBase`
            The potential object to compute the energy from.

        Returns
        -------
        E : :class:`~astropy.units.Quantity`
            The potential energy.
        """
        return potential.value(self.pos)

    def energy(self, potential):
        r"""
        The total energy *per unit mass* (e.g., kinetic + potential):

        Parameters
        ----------
        potential : `gala.potential.PotentialBase`
            The potential object to compute the energy from.

        Returns
        -------
        E : :class:`~astropy.units.Quantity`
            The total energy.
        """
        return self.kinetic_energy() + self.potential_energy(potential)

    def angular_momentum(self):
        r"""
        Compute the angular momentum for the phase-space positions contained
        in this object::

        .. math::

            \boldsymbol{{L}} = \boldsymbol{{q}} \times \boldsymbol{{p}}

        See :ref:`shape-conventions` for more information about the shapes of
        input and output objects.

        Returns
        -------
        L : :class:`~astropy.units.Quantity`
            Array of angular momentum vectors.

        Examples
        --------

            >>> import numpy as np
            >>> import astropy.units as u
            >>> pos = np.array([1., 0, 0]) * u.au
            >>> vel = np.array([0, 2*np.pi, 0]) * u.au/u.yr
            >>> w = {}(pos, vel)
            >>> w.angular_momentum()
            <Quantity [ 0.        , 0.        , 6.28318531] AU2 / yr>
        """.format(self.__class__.__name__)
        return np.cross(self.pos.value, self.vel.value, axis=0) * self.pos.unit * self.vel.unit

    # ------------------------------------------------------------------------
    # Misc. useful methods
    # ------------------------------------------------------------------------
    def plot(self, **kwargs):
        """
        Plot the positions in all projections. This is a thin wrapper around
        `~gala.dynamics.three_panel` -- the docstring for this function is
        included here.

        .. warning::

            This will currently fail for orbits with fewer than 3 dimensions.

        Parameters
        ----------
        relative_to : bool (optional)
            Plot the values relative to this value or values.
        autolim : bool (optional)
            Automatically set the plot limits to be something sensible.
        axes : array_like (optional)
            Array of matplotlib Axes objects.
        triangle : bool (optional)
            Make a triangle plot instead of plotting all projections in a single row.
        subplots_kwargs : dict (optional)
            Dictionary of kwargs passed to :func:`~matplotlib.pyplot.subplots`.
        labels : iterable (optional)
            List or iterable of axis labels as strings. They should correspond to the
            dimensions of the input orbit.
        **kwargs
            All other keyword arguments are passed to :func:`~matplotlib.pyplot.scatter`.
            You can pass in any of the usual style kwargs like ``color=...``,
            ``marker=...``, etc.

        Returns
        -------
        fig : `~matplotlib.Figure`

        """
        _label_unit = ''
        if self.pos.unit != u.dimensionless_unscaled:
            _label_unit = ' [{}]'.format(self.pos.unit)

        default_kwargs = {
            'marker': '.',
            'color': 'k',
            'labels': ('$x${}'.format(_label_unit),
                       '$y${}'.format(_label_unit),
                       '$z${}'.format(_label_unit))
        }

        for k,v in default_kwargs.items():
            kwargs[k] = kwargs.get(k, v)

        return three_panel(self.pos.value, **kwargs)

def combine(args):
    """
    Combine the input `PhaseSpacePosition` objects into a single object.

    Parameters
    ----------
    args : iterable
        Multiple instances of `PhaseSpacePosition`.

    Returns
    -------
    obj : `~gala.dynamics.PhaseSpacePosition`
        A single objct with positions and velocities stacked along the last axis.
    """

    ndim = None
    pos_unit = None
    vel_unit = None
    all_pos = []
    all_vel = []
    for x in args:
        if ndim is None:
            ndim = x.ndim
            pos_unit = x.pos.unit
            vel_unit = x.vel.unit
        else:
            if x.ndim != ndim:
                raise ValueError("All objects must have the same dimensionality.")

        pos = x.pos
        if pos.ndim < 2:
            pos = pos[...,np.newaxis]

        vel = x.vel
        if vel.ndim < 2:
            vel = vel[...,np.newaxis]

        all_pos.append(pos.to(pos_unit).value)
        all_vel.append(vel.to(vel_unit).value)

    all_pos = np.hstack(all_pos)*pos_unit
    all_vel = np.hstack(all_vel)*vel_unit

    return CartesianPhaseSpacePosition(pos=all_pos, vel=all_vel)




# coding: utf-8

from __future__ import division, print_function

__author__ = "adrn <adrn@astro.columbia.edu>"

# Third-party
import numpy as np

__all__ = ["cartesian_to_poincare_polar"]

def cartesian_to_poincare_polar(w):
    r"""
    Convert an array of 6D Cartesian positions to Poincaré
    symplectic polar coordinates. These are similar to cylindrical
    coordinates.

    Parameters
    ----------
    w : array_like
        Input array of 6D Cartesian phase-space positions. Should have
        shape ``(norbits,6)``.

    Returns
    -------
    new_w : :class:`~numpy.ndarray`
        Points represented in 6D Poincaré polar coordinates.

    """

    R = np.sqrt(w[...,0]**2 + w[...,1]**2)
    # phi = np.arctan2(w[...,1], w[...,0])
    phi = np.arctan2(w[...,0], w[...,1])

    vR = (w[...,0]*w[...,0+3] + w[...,1]*w[...,1+3]) / R
    vPhi = w[...,0]*w[...,1+3] - w[...,1]*w[...,0+3]

    # pg. 437, Papaphillipou & Laskar (1996)
    sqrt_2THETA = np.sqrt(np.abs(2*vPhi))
    pp_phi = sqrt_2THETA * np.cos(phi)
    pp_phidot = sqrt_2THETA * np.sin(phi)

    z = w[...,2]
    zdot = w[...,2+3]

    new_w = np.vstack((R.T, pp_phi.T, z.T,
                       vR.T, pp_phidot.T, zdot.T)).T
    return new_w



# coding: utf-8

""" Leapfrog integration. """

from __future__ import division, print_function

__author__ = "adrn <adrn@astro.columbia.edu>"

# Third-party
import numpy as np

# Project
from ..core import Integrator
from ..timespec import parse_time_specification

__all__ = ["LeapfrogIntegrator"]
class LeapfrogIntegrator(Integrator):
    r"""
    A symplectic, Leapfrog integrator.

    Given a function for computing time derivatives of the phase-space
    coordinates, this object computes the orbit at specified times.

    .. seealso::

        - http://en.wikipedia.org/wiki/Leapfrog_integration
        - http://ursa.as.arizona.edu/~rad/phys305/ODE_III/node11.html

    Naming convention for variables::

        im1 = i-1
        im1_2 = i-1/2
        ip1 = i+1
        ip1_2 = i+1/2

    Examples
    --------

    Using ``q`` as our coordinate variable and ``p`` as the conjugate
    momentum, we want to numerically solve for an orbit in the
    potential (Hamiltonian)

    .. math::

        \Phi &= \frac{1}{2}q^2\\
        H(q,p) &= \frac{1}{2}(p^2 + q^2)


    In this system,

    .. math::

        \dot{q} &= \frac{\partial \Phi}{\partial p} = p \\
        \dot{p} &= -\frac{\partial \Phi}{\partial q} = -q


    We will use the variable ``w`` to represent the full phase-space vector,
    :math:`w = (q,p)`. We define a function that computes the time derivates
    at any given time, ``t``, and phase-space position, ``w``::

        def F(t,w):
            dw = [w[1], -w[0]]
            return dw

    .. note::

        The force here is not time dependent, but this function always has
        to accept the independent variable (e.g., time) as the
        first argument.

    To create an integrator object, just pass this acceleration function in
    to the constructor, and then we can integrate orbits from a given vector
    of initial conditions::

        integrator = LeapfrogIntegrator(acceleration)
        times,ws = integrator.run(w0=[1.,0.], dt=0.1, n_steps=1000)

    .. note::

        When integrating a single vector of initial conditions, the return
        array will have 2 axes. In the above example, the returned array will
        have shape ``(2,1001)``. If an array of initial conditions are passed
        in, the return array will have 3 axes, where the last axis is for the
        individual orbits.

    Parameters
    ----------
    func : func
        A callable object that computes the phase-space time derivatives
        at a time and point in phase space.
    func_args : tuple (optional)
        Any extra arguments for the derivative function.
    func_units : `~gala.units.UnitSystem` (optional)
        If using units, this is the unit system assumed by the
        integrand function.

    """

    def step(self, t, x_im1, v_im1_2, dt):
        """
        Step forward the positions and velocities by the given timestep.

        Parameters
        ----------
        dt : numeric
            The timestep to move forward.
        """

        x_i = x_im1 + v_im1_2 * dt
        F_i = self.F(t, np.vstack((x_i,v_im1_2)), *self._func_args)
        a_i = F_i[self.ndim:]

        v_i = v_im1_2 + a_i * dt / 2
        v_ip1_2 = v_i + a_i * dt / 2

        return x_i, v_i, v_ip1_2

    def _init_v(self, t, w0, dt):
        """
        Leapfrog updates the velocities offset a half-step from the
        position updates. If we're given initial conditions aligned in
        time, e.g. the positions and velocities at the same 0th step,
        then we have to initially scoot the velocities forward by a half
        step to prime the integrator.

        Parameters
        ----------
        dt : numeric
            The first timestep.
        """

        # here is where we scoot the velocity at t=t1 to v(t+1/2)
        F0 = self.F(t.copy(), w0.copy(), *self._func_args)
        a0 = F0[self.ndim:]
        v_1_2 = w0[self.ndim:] + a0*dt/2.

        return v_1_2

    def run(self, w0, mmap=None, **time_spec):

        # generate the array of times
        times = parse_time_specification(self._func_units, **time_spec)
        n_steps = len(times) - 1
        _dt = times[1] - times[0]

        w0_obj, w0, ws = self._prepare_ws(w0, mmap, n_steps)
        x0 = w0[:self.ndim]
        v0 = w0[self.ndim:]

        if _dt < 0.:
            v0 *= -1.
            dt = np.abs(_dt)
        else:
            dt = _dt

        # prime the integrator so velocity is offset from coordinate by a
        #   half timestep
        v_im1_2 = self._init_v(times[0], w0, dt)
        x_im1 = x0

        ws[:,0] = w0
        for ii in range(1,n_steps+1):
            x_i, v_i, v_ip1_2 = self.step(times[ii], x_im1, v_im1_2, dt)
            ws[:self.ndim,ii,:] = x_i
            ws[self.ndim:,ii,:] = v_i
            x_im1, v_im1_2 = x_i, v_ip1_2

        if _dt < 0:
            ws[self.ndim:,...] *= -1.

        return self._handle_output(w0_obj, times, ws)



# coding: utf-8
"""
    Test conversions in core.py
"""

from __future__ import absolute_import, division, print_function

__author__ = "adrn <adrn@astro.columbia.edu>"

# Standard library
import os
import pytest
import numpy as np
import tempfile

# Third-party
import astropy.coordinates as coord
import astropy.units as u
from astropy.utils.data import get_pkg_data_filename

# This package
from ..core import *

def test_vgsr_to_vhel():
    filename = get_pkg_data_filename('idl_vgsr_vhel.txt')
    data = np.genfromtxt(filename, names=True, skip_header=2)

    # one row
    row = data[0]
    l = coord.Angle(row["lon"] * u.degree)
    b = coord.Angle(row["lat"] * u.degree)
    c = coord.Galactic(l, b)
    vgsr = row["vgsr"] * u.km/u.s
    vlsr = [row["vx"],row["vy"],row["vz"]]*u.km/u.s # this is right
    vcirc = row["vcirc"]*u.km/u.s

    vhel = vgsr_to_vhel(c, vgsr, vlsr=vlsr, vcirc=vcirc)
    np.testing.assert_almost_equal(vhel.value, row['vhelio'], decimal=4)

    # now check still get right answer passing in ICRS coordinates
    vhel = vgsr_to_vhel(c.transform_to(coord.ICRS), vgsr,
                        vlsr=vlsr, vcirc=vcirc)
    np.testing.assert_almost_equal(vhel.value, row['vhelio'], decimal=4)

    # all together now
    l = coord.Angle(data["lon"] * u.degree)
    b = coord.Angle(data["lat"] * u.degree)
    c = coord.Galactic(l, b)
    vgsr = data["vgsr"] * u.km/u.s
    vhel = vgsr_to_vhel(c, vgsr, vlsr=vlsr, vcirc=vcirc)
    np.testing.assert_almost_equal(vhel.value, data['vhelio'], decimal=4)

    # now check still get right answer passing in ICRS coordinates
    vhel = vgsr_to_vhel(c.transform_to(coord.ICRS), vgsr,
                        vlsr=vlsr, vcirc=vcirc)
    np.testing.assert_almost_equal(vhel.value, data['vhelio'], decimal=4)

def test_vgsr_to_vhel_misc():
    # make sure it works with longitude in 0-360 or -180-180
    l1 = coord.Angle(190.*u.deg)
    l2 = coord.Angle(-170.*u.deg)
    b = coord.Angle(30.*u.deg)

    c1 = coord.Galactic(l1, b)
    c2 = coord.Galactic(l2, b)

    vgsr = -110.*u.km/u.s
    vhel1 = vgsr_to_vhel(c1,vgsr)
    vhel2 = vgsr_to_vhel(c2,vgsr)

    np.testing.assert_almost_equal(vhel1.value, vhel2.value, decimal=9)

    # make sure throws error if tuple elements are not Quantities
    with pytest.raises(TypeError):
        vgsr_to_vhel(c1, vgsr.value)

def test_vhel_to_vgsr():
    filename = get_pkg_data_filename('idl_vgsr_vhel.txt')
    data = np.genfromtxt(filename, names=True, skip_header=2)

    # one row
    row = data[0]
    l = coord.Angle(row["lon"] * u.degree)
    b = coord.Angle(row["lat"] * u.degree)
    c = coord.Galactic(l, b)
    vhel = row["vhelio"] * u.km/u.s
    vlsr = [row["vx"],row["vy"],row["vz"]]*u.km/u.s # this is right
    vcirc = row["vcirc"]*u.km/u.s

    vgsr = vhel_to_vgsr(c, vhel, vlsr=vlsr, vcirc=vcirc)
    np.testing.assert_almost_equal(vgsr.value, row['vgsr'], decimal=4)

    # now check still get right answer passing in ICRS coordinates
    vgsr = vhel_to_vgsr(c.transform_to(coord.ICRS), vhel,
                        vlsr=vlsr, vcirc=vcirc)
    np.testing.assert_almost_equal(vgsr.value, row['vgsr'], decimal=4)

    # all together now
    l = coord.Angle(data["lon"] * u.degree)
    b = coord.Angle(data["lat"] * u.degree)
    c = coord.Galactic(l, b)
    vhel = data["vhelio"] * u.km/u.s
    vgsr = vhel_to_vgsr(c, vhel, vlsr=vlsr, vcirc=vcirc)
    np.testing.assert_almost_equal(vgsr.value, data['vgsr'], decimal=4)

    # now check still get right answer passing in ICRS coordinates
    vgsr = vhel_to_vgsr(c.transform_to(coord.ICRS), vhel,
                        vlsr=vlsr, vcirc=vcirc)
    np.testing.assert_almost_equal(vgsr.value, data['vgsr'], decimal=4)

def test_vhel_to_vgsr_misc():
    vhel = 110*u.km/u.s
    c1 = coord.Galactic(15*u.deg, -0.6*u.deg)

    # make sure throws error if tuple elements are not Quantities
    with pytest.raises(TypeError):
        vhel_to_vgsr(c1, vhel.value)

_txt = """# from: XHIP catalog
# ra dec HIPID l b dist pml pmb rv U V W
0.022010 20.036114      7 106.82021040 -41.22316218   57.56  -253.69  -138.84    8.30   71.7    2.1  -34.0
2.208349 40.494550    714 114.23363142 -21.65650026  249.00     5.57    -9.00  -11.78    0.1  -16.3   -5.5
3.126297 14.563522    999 108.98177530 -47.25067692   40.94   296.66  -141.05  -15.30  -44.5  -47.6   -7.3
"""
class TestVHelGalConvert(object):

    def setup(self):
        with tempfile.NamedTemporaryFile(mode='w+b') as temp:
            temp.write(_txt.encode('utf-8'))
            temp.flush()
            temp.seek(0)
            self.data = np.genfromtxt(temp, names=True, skip_header=1)

    def test_vhel_to_gal_single(self):

        # test a single entry
        row = self.data[0]
        c = coord.SkyCoord(ra=row['ra']*u.deg, dec=row['dec']*u.deg, distance=row['dist']*u.pc)
        pm = [row['pml'], row['pmb']]*u.mas/u.yr
        rv = row['rv']*u.km/u.s

        # stupid check
        vxyz_i = vhel_to_gal(c.icrs, pm=pm, rv=rv,
                             vcirc=0*u.km/u.s,
                             vlsr=[0.,0,0]*u.km/u.s)

        vxyz = vhel_to_gal(c.galactic, pm=pm, rv=rv,
                           vcirc=0*u.km/u.s,
                           vlsr=[0.,0,0]*u.km/u.s)

        assert vxyz_i.shape == vxyz.shape

        true_UVW = [row['U'],row['V'],row['W']]*u.km/u.s
        found_UVW = vxyz
        np.testing.assert_allclose(true_UVW.value, found_UVW.value, atol=1.)

        # some sanity checks - first, some convenience definitions
        g = coord.Galactic(l=0*u.deg, b=0*u.deg).transform_to(coord.ICRS)
        galcen_frame = coord.Galactocentric(galcen_ra=g.ra,
                                            galcen_dec=g.dec,
                                            z_sun=0*u.kpc)

        # --------------------------------------------------------------------
        # l = 0
        # without LSR and circular velocity
        c = coord.SkyCoord(ra=galcen_frame.galcen_ra, dec=galcen_frame.galcen_dec, distance=2*u.kpc)
        pm = [0., 0]*u.mas/u.yr
        rv = 20*u.km/u.s
        vxyz = vhel_to_gal(c.galactic, pm=pm, rv=rv,
                           vcirc=0*u.km/u.s,
                           vlsr=[0.,0,0]*u.km/u.s,
                           galactocentric_frame=galcen_frame)
        np.testing.assert_allclose(vxyz.to(u.km/u.s).value, [20,0,0.], atol=1E-12)

        # with LSR and circular velocity
        c = coord.SkyCoord(ra=galcen_frame.galcen_ra, dec=galcen_frame.galcen_dec, distance=2*u.kpc)
        pm = [0., 0]*u.mas/u.yr
        rv = 20*u.km/u.s
        vxyz = vhel_to_gal(c.galactic, pm=pm, rv=rv,
                           vcirc=200*u.km/u.s,
                           vlsr=[-20.,0,10]*u.km/u.s,
                           galactocentric_frame=galcen_frame)
        np.testing.assert_allclose(vxyz.to(u.km/u.s).value, [0,200,10], atol=1E-12)

        # l = 90
        # with LSR and circular velocity
        c = coord.SkyCoord(l=90*u.deg, b=0*u.deg, distance=2*u.kpc, frame=coord.Galactic)
        pm = [0., 0]*u.mas/u.yr
        rv = 20*u.km/u.s
        vxyz = vhel_to_gal(c.galactic, pm=pm, rv=rv,
                           vcirc=200*u.km/u.s,
                           vlsr=[-20.,0,10]*u.km/u.s,
                           galactocentric_frame=galcen_frame)
        np.testing.assert_allclose(vxyz.to(u.km/u.s).value, [-20,220,10], atol=1E-5)

        # l = 180
        # with LSR and circular velocity
        c = coord.SkyCoord(l=180*u.deg, b=0*u.deg, distance=2*u.kpc, frame=coord.Galactic)
        pm = [0., 0]*u.mas/u.yr
        rv = 20*u.km/u.s
        vxyz = vhel_to_gal(c.galactic, pm=pm, rv=rv,
                           vcirc=200*u.km/u.s,
                           vlsr=[-20.,0,10]*u.km/u.s,
                           galactocentric_frame=galcen_frame)
        np.testing.assert_allclose(vxyz.to(u.km/u.s).value, [-40,200,10], atol=1E-12)

        # l = 270
        # with LSR and circular velocity
        c = coord.SkyCoord(l=270*u.deg, b=0*u.deg, distance=2*u.kpc, frame=coord.Galactic)
        pm = [0., 0]*u.mas/u.yr
        rv = 20*u.km/u.s
        vxyz = vhel_to_gal(c.galactic, pm=pm, rv=rv,
                           vcirc=200*u.km/u.s,
                           vlsr=[-20.,0,10]*u.km/u.s,
                           galactocentric_frame=galcen_frame)
        np.testing.assert_allclose(vxyz.to(u.km/u.s).value, [-20,180,10], atol=1E-5)

    def test_vhel_to_gal_array(self):
        # test all together
        d = self.data
        c = coord.SkyCoord(ra=d['ra']*u.deg, dec=d['dec']*u.deg, distance=d['dist']*u.pc)
        pm = np.vstack((d['pml'], d['pmb']))*u.mas/u.yr
        rv = d['rv']*u.km/u.s

        # stupid check
        vxyz_i = vhel_to_gal(c.icrs, pm=pm, rv=rv,
                             vcirc=0*u.km/u.s,
                             vlsr=[0.,0,0]*u.km/u.s)
        vxyz = vhel_to_gal(c.galactic, pm=pm, rv=rv,
                           vcirc=0*u.km/u.s,
                           vlsr=[0.,0,0]*u.km/u.s)
        assert vxyz_i.shape == vxyz.shape

        # check values
        true_UVW = np.vstack((d['U'],d['V'],d['W']))*u.km/u.s
        found_UVW = vxyz
        np.testing.assert_allclose(true_UVW.value, found_UVW.value, atol=1.)

    def test_vgal_to_hel_single(self):

        # test a single entry
        row = self.data[0]
        c = coord.SkyCoord(ra=row['ra']*u.deg, dec=row['dec']*u.deg, distance=row['dist']*u.pc)
        pm = [row['pml'],row['pmb']]*u.mas/u.yr
        rv = row['rv']*u.km/u.s

        true_pmrv = (pm[0], pm[1], rv)
        vxyz = [row['U'],row['V'],row['W']]*u.km/u.s
        pmrv = vgal_to_hel(c.galactic, vxyz=vxyz,
                           vcirc=0.*u.km/u.s,
                           vlsr=[0.,0,0]*u.km/u.s)

        for i in range(3):
            np.testing.assert_allclose(pmrv[i].to(true_pmrv[i].unit).value,
                                       true_pmrv[i].value,
                                       atol=1.)

        # some sanity checks - first, some convenience definitions
        g = coord.Galactic(l=0*u.deg, b=0*u.deg).transform_to(coord.ICRS)
        frargs = dict(galcen_ra=g.ra,
                      galcen_dec=g.dec,
                      z_sun=0*u.kpc,
                      galcen_distance=8*u.kpc)
        galcen_frame = coord.Galactocentric(**frargs)

        # --------------------------------------------------------------------
        # l = 0
        # without LSR and circular velocity
        # c = coord.Galactocentric([6,0,0]*u.kpc,**frargs)
        c = coord.SkyCoord(l=0*u.deg, b=0*u.deg, distance=2*u.kpc, frame=coord.Galactic)
        vxyz = [20.,0,0]*u.km/u.s
        pmv = vgal_to_hel(c.galactic, vxyz,
                          vcirc=0*u.km/u.s,
                          vlsr=[0.,0,0]*u.km/u.s,
                          galactocentric_frame=galcen_frame)
        np.testing.assert_allclose(pmv[0].to(u.mas/u.yr).value, 0., atol=1E-12)
        np.testing.assert_allclose(pmv[1].to(u.mas/u.yr).value, 0., atol=1E-12)
        np.testing.assert_allclose(pmv[2].to(u.km/u.s).value, 20., atol=1E-12)

        # with LSR and circular velocity
        c = coord.SkyCoord(l=0*u.deg, b=0*u.deg, distance=2*u.kpc, frame=coord.Galactic)
        vxyz = [20.,0,0]*u.km/u.s
        pmv = vgal_to_hel(c.galactic, vxyz,
                          vcirc=-200*u.km/u.s,
                          vlsr=[0.,0,10]*u.km/u.s,
                          galactocentric_frame=galcen_frame)

        with u.set_enabled_equivalencies(u.dimensionless_angles()):
            np.testing.assert_allclose(pmv[0].to(u.mas/u.yr).value,
                                       ((200.*u.km/u.s)/(2*u.kpc)).to(u.mas/u.yr).value,
                                       atol=1E-12)
            np.testing.assert_allclose(pmv[1].to(u.mas/u.yr).value,
                                       ((-10.*u.km/u.s)/(2*u.kpc)).to(u.mas/u.yr).value,
                                       atol=1E-4)
        np.testing.assert_allclose(pmv[2].to(u.km/u.s).value, 20., atol=1E-12)

    def test_vgal_to_hel_array(self):
        # test all together
        d = self.data
        c = coord.SkyCoord(ra=d['ra']*u.deg, dec=d['dec']*u.deg, distance=d['dist']*u.pc)
        pm = np.vstack([d['pml'],d['pmb']])*u.mas/u.yr
        rv = d['rv']*u.km/u.s

        true_pmrv = (pm[0], pm[1], rv)
        vxyz = np.vstack((d['U'],d['V'],d['W']))*u.km/u.s
        pmrv = vgal_to_hel(c.galactic, vxyz=vxyz,
                           vcirc=0.*u.km/u.s,
                           vlsr=[0.,0,0]*u.km/u.s)

        for i in range(3):
            np.testing.assert_allclose(pmrv[i].to(true_pmrv[i].unit).value,
                                       true_pmrv[i].value,
                                       atol=1.)

    def test_roundtrip_icrs(self):
        np.random.seed(42)
        n = 100

        # yeahhhh, i know this isn't uniform on the sphere - shut up
        c = coord.SkyCoord(ra=np.random.uniform(0,360,n)*u.degree,
                           dec=np.random.uniform(-90,90,n)*u.degree,
                           distance=np.random.uniform(0.1,10.,n)*u.kpc)

        pm = np.random.uniform(-20,20,size=(2,n)) * u.mas/u.yr
        vr = np.random.normal(0., 75., size=n)*u.km/u.s
        mua,mud = pm  # initial

        # first to galactocentric
        vxyz = vhel_to_gal(c.icrs, pm=pm, rv=vr)

        # then back again, wooo
        pmv = vgal_to_hel(c.icrs, vxyz=vxyz)

        mua2,mud2 = pmv[:2]
        vr2 = pmv[2]

        np.testing.assert_allclose(mua.to(u.mas/u.yr).value, mua2.to(u.mas/u.yr).value, atol=1e-12)
        np.testing.assert_allclose(mud.to(u.mas/u.yr).value, mud2.to(u.mas/u.yr).value, atol=1e-12)
        np.testing.assert_allclose(vr.to(u.km/u.s).value, vr2.to(u.km/u.s).value, atol=1e-12)

    def test_roundtrip_gal(self):
        np.random.seed(42)
        n = 100

        # yeahhhh, i know this isn't uniform on the sphere - shut up
        c = coord.SkyCoord(ra=np.random.uniform(0,360,n)*u.degree,
                           dec=np.random.uniform(-90,90,n)*u.degree,
                           distance=np.random.uniform(0.1,10.,n)*u.kpc)

        pm = np.random.uniform(-20,20,size=(2,n)) * u.mas/u.yr
        vr = np.random.normal(0., 75., size=n)*u.km/u.s
        mul,mub = pm  # initial

        # first to galactocentric
        vxyz = vhel_to_gal(c.galactic, pm=pm, rv=vr)

        # then back again, wooo
        pmv = vgal_to_hel(c.galactic, vxyz=vxyz)

        mul2,mub2 = pmv[:2]
        vr2 = pmv[2]

        np.testing.assert_allclose(mul.to(u.mas/u.yr).value, mul2.to(u.mas/u.yr).value, rtol=1E-5, atol=1e-12)
        np.testing.assert_allclose(mub.to(u.mas/u.yr).value, mub2.to(u.mas/u.yr).value, rtol=1E-5, atol=1e-12)
        np.testing.assert_allclose(vr.to(u.km/u.s).value, vr2.to(u.km/u.s).value, rtol=1E-5, atol=1e-12)



from __future__ import absolute_import
from distutils.core import Extension
from astropy_helpers import setup_helpers

def get_extensions():
    exts = []

    # malloc
    mac_incl_path = "/usr/include/malloc"

    cfg = setup_helpers.DistutilsExtensionArgs()
    cfg['include_dirs'].append('numpy')
    cfg['include_dirs'].append(mac_incl_path)
    cfg['include_dirs'].append('gala/potential')
    cfg['extra_compile_args'].append('--std=gnu99')
    cfg['sources'].append('gala/potential/cpotential.pyx')
    cfg['sources'].append('gala/potential/src/cpotential.c')
    exts.append(Extension('gala.potential.cpotential', **cfg))

    cfg = setup_helpers.DistutilsExtensionArgs()
    cfg['include_dirs'].append('numpy')
    cfg['include_dirs'].append(mac_incl_path)
    cfg['include_dirs'].append('gala/potential')
    cfg['extra_compile_args'].append('--std=gnu99')
    cfg['sources'].append('gala/potential/builtin/cybuiltin.pyx')
    cfg['sources'].append('gala/potential/builtin/src/_cbuiltin.c')
    cfg['sources'].append('gala/potential/src/cpotential.c')
    exts.append(Extension('gala.potential.builtin.cybuiltin', **cfg))

    return exts

def get_package_data():
    return {'gala.potential': ['*.h', '*.pyx', '*.pxd', '*/*.pyx', '*/*.pxd',
                               'builtin/src/*.h', 'src/*.h',
                               'builtin/src/*.c', 'src/*.c',
                               'tests/*.yml']}



# coding: utf-8
# cython: boundscheck=False
# cython: debug=False
# cython: nonecheck=False
# cython: cdivision=True
# cython: wraparound=False
# cython: profile=False

""" Coordinate help for generating mock streams. """

from __future__ import division, print_function

__author__ = "adrn <adrn@astro.columbia.edu>"

from libc.math cimport M_PI

cdef extern from "math.h":
    double fabs(double x) nogil
    double sqrt(double x) nogil
    double cos(double x) nogil
    double sin(double x) nogil
    double atan2(double y, double x) nogil
    double fmod(double y, double x) nogil

cdef void sat_rotation_matrix(double *w, # in
                              double *R): # out
    cdef:
        double x1_norm, x2_norm, x3_norm = 0.
        unsigned int i
        double *x1 = [0.,0.,0.]
        double *x2 = [0.,0.,0.]
        double *x3 = [0.,0.,0.]

    x1[0] = w[0]
    x1[1] = w[1]
    x1[2] = w[2]

    x3[0] = x1[1]*w[2+3] - x1[2]*w[1+3]
    x3[1] = x1[2]*w[0+3] - x1[0]*w[2+3]
    x3[2] = x1[0]*w[1+3] - x1[1]*w[0+3]

    x2[0] = -x1[1]*x3[2] + x1[2]*x3[1]
    x2[1] = -x1[2]*x3[0] + x1[0]*x3[2]
    x2[2] = -x1[0]*x3[1] + x1[1]*x3[0]

    x1_norm = sqrt(x1[0]*x1[0] + x1[1]*x1[1] + x1[2]*x1[2])
    x2_norm = sqrt(x2[0]*x2[0] + x2[1]*x2[1] + x2[2]*x2[2])
    x3_norm = sqrt(x3[0]*x3[0] + x3[1]*x3[1] + x3[2]*x3[2])

    for i in range(3):
        x1[i] /= x1_norm
        x2[i] /= x2_norm
        x3[i] /= x3_norm

    R[0] = x1[0]
    R[1] = x1[1]
    R[2] = x1[2]
    R[3] = x2[0]
    R[4] = x2[1]
    R[5] = x2[2]
    R[6] = x3[0]
    R[7] = x3[1]
    R[8] = x3[2]

cdef void to_sat_coords(double *w, double *R, # in
                        double *w_prime): # out
    # Translate to be centered on progenitor
    cdef int i

    # Project into new basis
    w_prime[0] = w[0]*R[0] + w[1]*R[1] + w[2]*R[2]
    w_prime[1] = w[0]*R[3] + w[1]*R[4] + w[2]*R[5]
    w_prime[2] = w[0]*R[6] + w[1]*R[7] + w[2]*R[8]

    w_prime[3] = w[3]*R[0] + w[4]*R[1] + w[5]*R[2]
    w_prime[4] = w[3]*R[3] + w[4]*R[4] + w[5]*R[5]
    w_prime[5] = w[3]*R[6] + w[4]*R[7] + w[5]*R[8]

cdef void from_sat_coords(double *w_prime, double *R, # in
                          double *w): # out
    cdef int i

    # Project back from sat plane
    w[0] = w_prime[0]*R[0] + w_prime[1]*R[3] + w_prime[2]*R[6]
    w[1] = w_prime[0]*R[1] + w_prime[1]*R[4] + w_prime[2]*R[7]
    w[2] = w_prime[0]*R[2] + w_prime[1]*R[5] + w_prime[2]*R[8]

    w[3] = w_prime[3]*R[0] + w_prime[4]*R[3] + w_prime[5]*R[6]
    w[4] = w_prime[3]*R[1] + w_prime[4]*R[4] + w_prime[5]*R[7]
    w[5] = w_prime[3]*R[2] + w_prime[4]*R[5] + w_prime[5]*R[8]

# ---------------------------------------------------------------------

cdef void car_to_cyl(double *w, # in
                     double *cyl): # out
    cdef:
        double R = sqrt(w[0]*w[0] + w[1]*w[1])
        double phi = atan2(w[1], w[0])
        double vR = (w[0]*w[3] + w[1]*w[4]) / R
        double vphi = (w[0]*w[4] - w[3]*w[1]) / R

    cyl[0] = R
    if phi < 0:
        phi = phi + 2*M_PI
    cyl[1] = phi
    cyl[2] = w[2]

    cyl[3] = vR
    cyl[4] = vphi
    cyl[5] = w[5]

cdef void cyl_to_car(double *cyl, # in
                     double *w): # out
    w[0] = cyl[0] * cos(cyl[1])
    w[1] = cyl[0] * sin(cyl[1])
    w[2] = cyl[2]

    w[3] = cyl[3] * cos(cyl[1]) - cyl[4] * sin(cyl[1])
    w[4] = cyl[3] * sin(cyl[1]) + cyl[4] * cos(cyl[1])
    w[5] = cyl[5]

# ---------------------------------------------------------------------
# Tests
#

cpdef _test_sat_rotation_matrix():
    import numpy as np
    np.random.seed(42)
    n = 1024

    cdef:
        double[::1] w = np.zeros(6)
        double[::1] wrot = np.zeros(6)
        double[::1] w2 = np.zeros(6)
        double[:,::1] R = np.zeros((3,3))
        unsigned int i, j

    for i in range(n):
        w = np.random.uniform(size=6)
        sat_rotation_matrix(&w[0], &R[0,0])

        x = np.array(R).dot(np.array(w)[:3])
        assert x[0] > 0
        assert np.allclose(x[1], 0)
        assert np.allclose(x[2], 0)

        v = np.array(R).dot(np.array(w)[3:])
        assert np.allclose(v[2], 0)
        for j in range(3):
            wrot[j] = x[j]
            wrot[j+3] = v[j]

        x2 = np.array(R.T).dot(np.array(wrot)[:3])
        v2 = np.array(R.T).dot(np.array(wrot)[3:])
        for j in range(3):
            w2[j] = x2[j]
            w2[j+3] = v2[j]

        for j in range(6):
            assert np.allclose(w[j], w2[j])

cpdef _test_to_sat_coords_roundtrip():
    import numpy as np
    np.random.seed(42)
    n = 1024

    cdef:
        double[:,::1] w = np.random.uniform(size=(n,6))
        double[:,::1] w_sat = np.random.uniform(size=(n,6))
        double[:,::1] R = np.zeros((3,3))

        double[::1] w_prime = np.zeros(6)
        double[::1] w2 = np.zeros(6)

        unsigned int i, j

    for i in range(n):
        sat_rotation_matrix(&w_sat[i,0], &R[0,0])
        to_sat_coords(&w[i,0], &R[0,0], &w_prime[0])
        from_sat_coords(&w_prime[0], &R[0,0], &w2[0])

        for j in range(6):
            assert np.allclose(w[i,j], w2[j])

cpdef _test_car_to_cyl_roundtrip():
    import numpy as np
    np.random.seed(42)
    n = 1024

    cdef:
        double[:,::1] w = np.random.uniform(-10,10,size=(n,6))
        double[::1] cyl = np.zeros(6)
        double[::1] w2 = np.zeros(6)

        unsigned int i, j

    for i in range(n):
        car_to_cyl(&w[i,0], &cyl[0])
        cyl_to_car(&cyl[0], &w2[0])
        for j in range(6):
            assert np.allclose(w[i,j], w2[j])

cpdef _test_cyl_to_car_roundtrip():
    import numpy as np
    # np.random.seed(42)
    n = 1024

    cdef:
        double[:,::1] cyl = np.random.uniform(0,2*np.pi,size=(n,6))
        double[::1] w = np.zeros(6)
        double[::1] cyl2 = np.zeros(6)

        unsigned int i, j

    for i in range(n):
        cyl_to_car(&cyl[i,0], &w[0])
        car_to_cyl(&w[0], &cyl2[0])
        for j in range(6):
            # assert np.allclose(cyl[i,j], cyl2[j])
            if not np.allclose(cyl[i,j], cyl2[j]):
                print(i,j,cyl[i,j], cyl2[j])
        print()


# cdef void car_to_sph(double *xyz, double *sph):
#     # TODO: note this isn't consistent with the velocity transform because of theta
#     # get out spherical components
#     cdef:
#         double d = sqrt(xyz[0]*xyz[0]+xyz[1]*xyz[1]+xyz[2]*xyz[2])
#         double phi = atan2(xyz[1], xyz[0])
#         double theta = acos(xyz[2] / d)

#     sph[0] = d
#     sph[1] = phi
#     sph[2] = theta

# cdef void sph_to_car(double *sph, double *xyz):
#     # TODO: note this isn't consistent with the velocity transform because of theta
#     # get out spherical components
#     xyz[0] = sph[0] * cos(sph[1]) * sin(sph[2])
#     xyz[1] = sph[0] * sin(sph[1]) * sin(sph[2])
#     xyz[2] = sph[0] * cos(sph[2])

# cdef void v_car_to_sph(double *xyz, double *vxyz, double *vsph):
#     # get out spherical components
#     cdef:
#         double d = sqrt(xyz[0]*xyz[0]+xyz[1]*xyz[1]+xyz[2]*xyz[2])
#         double dxy = sqrt(xyz[0]*xyz[0] + xyz[1]*xyz[1])

#         double vr = (xyz[0]*vxyz[0]+xyz[1]*vxyz[1]+xyz[2]*vxyz[2]) / d

#         double mu_lon = (xyz[0]*vxyz[1] - vxyz[0]*xyz[1]) / (dxy*dxy)
#         double vlon = mu_lon * dxy # cos(lat)

#         double mu_lat = (xyz[2]*(xyz[0]*vxyz[0] + xyz[1]*vxyz[1]) - dxy*dxy*vxyz[2]) / (d*d*dxy)
#         double vlat = -mu_lat * d

#     vsph[0] = vr
#     vsph[1] = vlon
#     vsph[2] = vlat

# cdef void v_sph_to_car(double *xyz, double *vsph, double *vxyz):
#     # get out spherical components
#     cdef:
#         double d = sqrt(xyz[0]*xyz[0]+xyz[1]*xyz[1]+xyz[2]*xyz[2])
#         double dxy = sqrt(xyz[0]*xyz[0] + xyz[1]*xyz[1])

#     vxyz[0] = vsph[0]*xyz[0]/dxy*dxy/d - xyz[1]/dxy*vsph[1] - xyz[0]/dxy*xyz[2]/d*vsph[2]
#     vxyz[1] = vsph[0]*xyz[1]/dxy*dxy/d + xyz[0]/dxy*vsph[1] - xyz[1]/dxy*xyz[2]/d*vsph[2]
#     vxyz[2] = vsph[0]*xyz[2]/d + dxy/d*vsph[2]




# coding: utf-8
# cython: boundscheck=False
# cython: nonecheck=False
# cython: cdivision=True
# cython: wraparound=False
# cython: profile=False

""" Built-in potentials implemented in Cython """

from __future__ import division, print_function

__author__ = "adrn <adrn@astro.columbia.edu>"

# Standard library
from collections import OrderedDict

# Third-party
from astropy.extern import six
from astropy.utils import InheritDocstrings
from astropy.coordinates.angles import rotation_matrix
from astropy.constants import G
import astropy.units as u
import numpy as np
cimport numpy as np
np.import_array()

# Project
from ..core import CompositePotential
from ..cpotential import CPotentialBase
from ..cpotential cimport CPotentialWrapper
from ...units import DimensionlessUnitSystem

cdef extern from "src/cpotential.h":
    enum:
        MAX_N_COMPONENTS = 16

    ctypedef double (*densityfunc)(double t, double *pars, double *q) nogil
    ctypedef double (*valuefunc)(double t, double *pars, double *q) nogil
    ctypedef void (*gradientfunc)(double t, double *pars, double *q, double *grad) nogil

    ctypedef struct CPotential:
        int n_components
        int n_dim
        densityfunc density[MAX_N_COMPONENTS]
        valuefunc value[MAX_N_COMPONENTS]
        gradientfunc gradient[MAX_N_COMPONENTS]
        int n_params[MAX_N_COMPONENTS]
        double *parameters[MAX_N_COMPONENTS]

    double c_value(CPotential *p, double t, double *q) nogil
    double c_density(CPotential *p, double t, double *q) nogil
    void c_gradient(CPotential *p, double t, double *q, double *grad) nogil

cdef extern from "src/_cbuiltin.h":
    double nan_density(double t, double *pars, double *q) nogil

    double henon_heiles_value(double t, double *pars, double *q) nogil
    void henon_heiles_gradient(double t, double *pars, double *q, double *grad) nogil

    double kepler_value(double t, double *pars, double *q) nogil
    void kepler_gradient(double t, double *pars, double *q, double *grad) nogil

    double isochrone_value(double t, double *pars, double *q) nogil
    void isochrone_gradient(double t, double *pars, double *q, double *grad) nogil
    double isochrone_density(double t, double *pars, double *q) nogil

    double hernquist_value(double t, double *pars, double *q) nogil
    void hernquist_gradient(double t, double *pars, double *q, double *grad) nogil
    double hernquist_density(double t, double *pars, double *q) nogil

    double plummer_value(double t, double *pars, double *q) nogil
    void plummer_gradient(double t, double *pars, double *q, double *grad) nogil
    double plummer_density(double t, double *pars, double *q) nogil

    double jaffe_value(double t, double *pars, double *q) nogil
    void jaffe_gradient(double t, double *pars, double *q, double *grad) nogil
    double jaffe_density(double t, double *pars, double *q) nogil

    double stone_value(double t, double *pars, double *q) nogil
    void stone_gradient(double t, double *pars, double *q, double *grad) nogil
    double stone_density(double t, double *pars, double *q) nogil

    double sphericalnfw_value(double t, double *pars, double *q) nogil
    void sphericalnfw_gradient(double t, double *pars, double *q, double *grad) nogil
    double sphericalnfw_density(double t, double *pars, double *q) nogil

    double flattenednfw_value(double t, double *pars, double *q) nogil
    void flattenednfw_gradient(double t, double *pars, double *q, double *grad) nogil
    double flattenednfw_density(double t, double *pars, double *q) nogil

    double miyamotonagai_value(double t, double *pars, double *q) nogil
    void miyamotonagai_gradient(double t, double *pars, double *q, double *grad) nogil
    double miyamotonagai_density(double t, double *pars, double *q) nogil

    double leesuto_value(double t, double *pars, double *q) nogil
    void leesuto_gradient(double t, double *pars, double *q, double *grad) nogil
    double leesuto_density(double t, double *pars, double *q) nogil

    double logarithmic_value(double t, double *pars, double *q) nogil
    void logarithmic_gradient(double t, double *pars, double *q, double *grad) nogil

    double rotating_logarithmic_value(double t, double *pars, double *q) nogil
    void rotating_logarithmic_gradient(double t, double *pars, double *q, double *grad) nogil

__all__ = ['HenonHeilesPotential', # Misc. potentials
           'KeplerPotential', 'HernquistPotential', 'IsochronePotential', 'PlummerPotential',
           'JaffePotential', 'SphericalNFWPotential', 'StonePotential', # Spherical models
           'MiyamotoNagaiPotential', 'FlattenedNFWPotential', # Flattened models
           'LeeSutoTriaxialNFWPotential', 'LogarithmicPotential', # Triaxial models
           'CCompositePotential']

# ============================================================================

cdef class HenonHeilesWrapper(CPotentialWrapper):

    def __init__(self, G, *args):
        cdef CPotential cp

        # This is the only code that needs to change per-potential
        cp.value[0] = <valuefunc>(henon_heiles_value)
        cp.density[0] = <densityfunc>(nan_density)
        cp.gradient[0] = <gradientfunc>(henon_heiles_gradient)
        self._params = np.array([G], dtype=np.float64)
        # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

        cp.n_components = 1
        self._n_params = np.array([len(self._params)], dtype=np.int32)
        cp.n_params = &(self._n_params[0])
        cp.parameters[0] = &(self._params[0])
        cp.n_dim = 2
        self.cpotential = cp

class HenonHeilesPotential(CPotentialBase):
    r"""
    HenonHeilesPotential(units=None)

    The Hénon-Heiles potential.

    .. math::

        \Phi(x,y) = \frac{1}{2}(x^2 + y^2 + 2x^2 y - \frac{2}{3}y^3)

    Parameters
    ----------
    units : `~gala.units.UnitSystem` (optional)
        Set of non-reducable units that specify (at minimum) the
        length, mass, time, and angle units.

    """
    def __init__(self, units=None):
        parameters = OrderedDict()
        super(HenonHeilesPotential, self).__init__(parameters=parameters,
                                                   units=units)

# ============================================================================

cdef class KeplerWrapper(CPotentialWrapper):

    def __init__(self, G, parameters):
        cdef CPotential cp

        # This is the only code that needs to change per-potential
        cp.value[0] = <valuefunc>(kepler_value)
        cp.density[0] = <densityfunc>(nan_density)
        cp.gradient[0] = <gradientfunc>(kepler_gradient)
        # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

        cp.n_components = 1
        self._params = np.array([G] + list(parameters), dtype=np.float64)
        self._n_params = np.array([len(self._params)], dtype=np.int32)
        cp.n_params = &(self._n_params[0])
        cp.parameters[0] = &(self._params[0])
        cp.n_dim = 3
        self.cpotential = cp

class KeplerPotential(CPotentialBase):
    r"""
    KeplerPotential(m, units)

    The Kepler potential for a point mass.

    .. math::

        \Phi(r) = -\frac{Gm}{r}

    Parameters
    ----------
    m : numeric
        Mass.

    Parameters
    ----------
    units : `~gala.units.UnitSystem` (optional)
        Set of non-reducable units that specify (at minimum) the
        length, mass, time, and angle units.

    """
    def __init__(self, m, units):
        parameters = OrderedDict()
        parameters['m'] = m
        super(KeplerPotential, self).__init__(parameters=parameters,
                                              units=units)

# ============================================================================

cdef class IsochroneWrapper(CPotentialWrapper):

    def __init__(self, G, parameters):
        cdef CPotential cp

        # This is the only code that needs to change per-potential
        cp.value[0] = <valuefunc>(isochrone_value)
        cp.density[0] = <densityfunc>(isochrone_density)
        cp.gradient[0] = <gradientfunc>(isochrone_gradient)
        # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

        cp.n_components = 1
        self._params = np.array([G] + list(parameters), dtype=np.float64)
        self._n_params = np.array([len(self._params)], dtype=np.int32)
        cp.n_params = &(self._n_params[0])
        cp.parameters[0] = &(self._params[0])
        cp.n_dim = 3
        self.cpotential = cp

class IsochronePotential(CPotentialBase):
    r"""
    IsochronePotential(m, b, units)

    The Isochrone potential.

    .. math::

        \Phi = -\frac{GM}{\sqrt{r^2+b^2} + b}

    Parameters
    ----------
    m : numeric
        Mass.
    b : numeric
        Core concentration.
    units : `~gala.units.UnitSystem` (optional)
        Set of non-reducable units that specify (at minimum) the
        length, mass, time, and angle units.

    """
    def __init__(self, m, b, units):
        parameters = OrderedDict()
        parameters['m'] = m
        parameters['b'] = b
        super(IsochronePotential, self).__init__(parameters=parameters,
                                                 units=units)

    def action_angle(self, w):
        """
        Transform the input cartesian position and velocity to action-angle
        coordinates the Isochrone potential. See Section 3.5.2 in
        Binney & Tremaine (2008), and be aware of the errata entry for
        Eq. 3.225.

        This transformation is analytic and can be used as a "toy potential"
        in the Sanders & Binney 2014 formalism for computing action-angle
        coordinates in _any_ potential.

        Adapted from Jason Sanders' code
        `here <https://github.com/jlsanders/genfunc>`_.

        Parameters
        ----------
        w : :class:`gala.dynamics.CartesianPhaseSpacePosition`, :class:`gala.dynamics.CartesianOrbit`
            The positions or orbit to compute the actions, angles, and frequencies at.
        """
        from ...dynamics.analyticactionangle import isochrone_to_aa
        return isochrone_to_aa(w, self)

    # def phase_space(self, actions, angles):
    #     """
    #     Transform the input actions and angles to ordinary phase space (position
    #     and velocity) in cartesian coordinates. See Section 3.5.2 in
    #     Binney & Tremaine (2008), and be aware of the errata entry for
    #     Eq. 3.225.

    #     Parameters
    #     ----------
    #     actions : array_like
    #     angles : array_like
    #     """
    #     from ...dynamics.analyticactionangle import isochrone_aa_to_xv
    #     return isochrone_aa_to_xv(actions, angles, self)

# ============================================================================

cdef class HernquistWrapper(CPotentialWrapper):

    def __init__(self, G, parameters):
        cdef CPotential cp

        # This is the only code that needs to change per-potential
        cp.value[0] = <valuefunc>(hernquist_value)
        cp.density[0] = <densityfunc>(hernquist_density)
        cp.gradient[0] = <gradientfunc>(hernquist_gradient)
        # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

        cp.n_components = 1
        self._params = np.array([G] + list(parameters), dtype=np.float64)
        self._n_params = np.array([len(self._params)], dtype=np.int32)
        cp.n_params = &(self._n_params[0])
        cp.parameters[0] = &(self._params[0])
        cp.n_dim = 3
        self.cpotential = cp

class HernquistPotential(CPotentialBase):
    r"""
    HernquistPotential(m, c, units)

    Hernquist potential for a spheroid.

    .. math::

        \Phi(r) = -\frac{G M}{r + c}

    See: http://adsabs.harvard.edu/abs/1990ApJ...356..359H

    Parameters
    ----------
    m : numeric
        Mass.
    c : numeric
        Core concentration.
    units : `~gala.units.UnitSystem` (optional)
        Set of non-reducable units that specify (at minimum) the
        length, mass, time, and angle units.

    """
    def __init__(self, m, c, units):
        parameters = OrderedDict()
        parameters['m'] = m
        parameters['c'] = c
        super(HernquistPotential, self).__init__(parameters=parameters,
                                                 units=units)

# ============================================================================

cdef class PlummerWrapper(CPotentialWrapper):

    def __init__(self, G, parameters):
        cdef CPotential cp

        # This is the only code that needs to change per-potential
        cp.value[0] = <valuefunc>(plummer_value)
        cp.density[0] = <densityfunc>(plummer_density)
        cp.gradient[0] = <gradientfunc>(plummer_gradient)
        # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

        cp.n_components = 1
        self._params = np.array([G] + list(parameters), dtype=np.float64)
        self._n_params = np.array([len(self._params)], dtype=np.int32)
        cp.n_params = &(self._n_params[0])
        cp.parameters[0] = &(self._params[0])
        cp.n_dim = 3
        self.cpotential = cp

class PlummerPotential(CPotentialBase):
    r"""
    PlummerPotential(m, b, units)

    Plummer potential for a spheroid.

    .. math::

        \Phi(r) = -\frac{G M}{\sqrt{r^2 + b^2}}

    Parameters
    ----------
    m : numeric
       Mass.
    b : numeric
        Core concentration.
    units : `~gala.units.UnitSystem` (optional)
        Set of non-reducable units that specify (at minimum) the
        length, mass, time, and angle units.

    """
    def __init__(self, m, b, units):
        parameters = OrderedDict()
        parameters['m'] = m
        parameters['b'] = b
        super(PlummerPotential, self).__init__(parameters=parameters,
                                               units=units)

# ============================================================================

cdef class JaffeWrapper(CPotentialWrapper):

    def __init__(self, G, parameters):
        cdef CPotential cp

        # This is the only code that needs to change per-potential
        cp.value[0] = <valuefunc>(jaffe_value)
        cp.density[0] = <densityfunc>(jaffe_density)
        cp.gradient[0] = <gradientfunc>(jaffe_gradient)
        # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

        cp.n_components = 1
        self._params = np.array([G] + list(parameters), dtype=np.float64)
        self._n_params = np.array([len(self._params)], dtype=np.int32)
        cp.n_params = &(self._n_params[0])
        cp.parameters[0] = &(self._params[0])
        cp.n_dim = 3
        self.cpotential = cp

class JaffePotential(CPotentialBase):
    r"""
    JaffePotential(m, c, units)

    Jaffe potential for a spheroid.

    .. math::

        \Phi(r) = \frac{G M}{c} \ln(\frac{r}{r + c})

    Parameters
    ----------
    m : numeric
        Mass.
    c : numeric
        Core concentration.
    units : `~gala.units.UnitSystem` (optional)
        Set of non-reducable units that specify (at minimum) the
        length, mass, time, and angle units.

    """
    def __init__(self, m, c, units):
        parameters = OrderedDict()
        parameters['m'] = m
        parameters['c'] = c
        super(JaffePotential, self).__init__(parameters=parameters,
                                             units=units)

# ============================================================================

cdef class StoneWrapper(CPotentialWrapper):

    def __init__(self, G, parameters):
        cdef CPotential cp

        # This is the only code that needs to change per-potential
        cp.value[0] = <valuefunc>(stone_value)
        cp.density[0] = <densityfunc>(stone_density)
        cp.gradient[0] = <gradientfunc>(stone_gradient)
        # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

        cp.n_components = 1
        self._params = np.array([G] + list(parameters), dtype=np.float64)
        self._n_params = np.array([len(self._params)], dtype=np.int32)
        cp.n_params = &(self._n_params[0])
        cp.parameters[0] = &(self._params[0])
        cp.n_dim = 3
        self.cpotential = cp

class StonePotential(CPotentialBase):
    r"""
    StonePotential(m, r_c, r_h, units)

    Stone potential from `Stone & Ostriker (2015) <http://dx.doi.org/10.1088/2041-8205/806/2/L28>`_.

    .. math::

        \Phi(r) = -\frac{2 G M}{\pi(r_h - r_c)}\left[ \frac{\arctan(r/r_h)}{r/r_h} - \frac{\arctan(r/r_c)}{r/r_c} + \frac{1}{2}\ln\left(\frac{r^2+r_h^2}{r^2+r_c^2}\right)\right]

    Parameters
    ----------
    m_tot : numeric
        Total mass.
    r_c : numeric
        Core radius.
    r_h : numeric
        Halo radius.
    units : `~gala.units.UnitSystem` (optional)
        Set of non-reducable units that specify (at minimum) the
        length, mass, time, and angle units.

    """
    def __init__(self, m, r_c, r_h, units):
        parameters = OrderedDict()
        parameters['m'] = m
        parameters['r_c'] = r_c
        parameters['r_h'] = r_h
        super(StonePotential, self).__init__(parameters=parameters,
                                             units=units)

# ============================================================================

cdef class SphericalNFWWrapper(CPotentialWrapper):

    def __init__(self, G, parameters):
        cdef CPotential cp

        # This is the only code that needs to change per-potential
        cp.value[0] = <valuefunc>(sphericalnfw_value)
        cp.density[0] = <densityfunc>(sphericalnfw_density)
        cp.gradient[0] = <gradientfunc>(sphericalnfw_gradient)
        # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

        cp.n_components = 1
        self._params = np.array([G] + list(parameters), dtype=np.float64)
        self._n_params = np.array([len(self._params)], dtype=np.int32)
        cp.n_params = &(self._n_params[0])
        cp.parameters[0] = &(self._params[0])
        cp.n_dim = 3
        self.cpotential = cp

class SphericalNFWPotential(CPotentialBase):
    r"""
    SphericalNFWPotential(v_c, r_s, units)

    Spherical NFW potential. Separate from the triaxial potential below to
    optimize for speed. Much faster than computing the triaxial case.

    .. math::

        \Phi(r) = -\frac{v_c^2}{\sqrt{\ln 2 - \frac{1}{2}}} \frac{\ln(1 + r/r_s)}{r/r_s}

    Parameters
    ----------
    v_c : numeric
        Circular velocity at the scale radius.
    r_s : numeric
        Scale radius.
    units : `~gala.units.UnitSystem` (optional)
        Set of non-reducable units that specify (at minimum) the
        length, mass, time, and angle units.

    """
    def __init__(self, v_c, r_s, units):
        parameters = OrderedDict()
        parameters['v_c'] = v_c
        parameters['r_s'] = r_s
        super(SphericalNFWPotential, self).__init__(parameters=parameters,
                                                    units=units)

# ============================================================================

cdef class MiyamotoNagaiWrapper(CPotentialWrapper):

    def __init__(self, G, parameters):
        cdef CPotential cp

        # This is the only code that needs to change per-potential
        cp.value[0] = <valuefunc>(miyamotonagai_value)
        cp.density[0] = <densityfunc>(miyamotonagai_density)
        cp.gradient[0] = <gradientfunc>(miyamotonagai_gradient)
        # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

        cp.n_components = 1
        self._params = np.array([G] + list(parameters), dtype=np.float64)
        self._n_params = np.array([len(self._params)], dtype=np.int32)
        cp.n_params = &(self._n_params[0])
        cp.parameters[0] = &(self._params[0])
        cp.n_dim = 3
        self.cpotential = cp

class MiyamotoNagaiPotential(CPotentialBase):
    r"""
    MiyamotoNagaiPotential(m, a, b, units)

    Miyamoto-Nagai potential for a flattened mass distribution.

    .. math::

        \Phi(R,z) = -\frac{G M}{\sqrt{R^2 + (a + \sqrt{z^2 + b^2})^2}}

    See: http://adsabs.harvard.edu/abs/1975PASJ...27..533M

    Parameters
    ----------
    m : numeric
        Mass.
    a : numeric
        Scale length.
    b : numeric
        Scare height.
    units : `~gala.units.UnitSystem` (optional)
        Set of non-reducable units that specify (at minimum) the
        length, mass, time, and angle units.

    """
    def __init__(self, m, a, b, units):
        parameters = OrderedDict()
        parameters['m'] = m
        parameters['a'] = a
        parameters['b'] = b
        super(MiyamotoNagaiPotential, self).__init__(parameters=parameters,
                                                     units=units)

# ============================================================================

cdef class FlattenedNFWWrapper(CPotentialWrapper):

    def __init__(self, G, parameters):
        cdef CPotential cp

        # This is the only code that needs to change per-potential
        cp.value[0] = <valuefunc>(flattenednfw_value)
        cp.density[0] = <densityfunc>(flattenednfw_density)
        cp.gradient[0] = <gradientfunc>(flattenednfw_gradient)
        # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

        cp.n_components = 1
        self._params = np.array([G] + list(parameters), dtype=np.float64)
        self._n_params = np.array([len(self._params)], dtype=np.int32)
        cp.n_params = &(self._n_params[0])
        cp.parameters[0] = &(self._params[0])
        cp.n_dim = 3
        self.cpotential = cp

class FlattenedNFWPotential(CPotentialBase):
    r"""
    FlattenedNFWPotential(v_c, r_s, q_z, units)

    Flattened NFW potential. Separate from the triaxial potential below to
    optimize for speed. Much faster than computing the triaxial case.

    .. math::

        \Phi(r) = -\frac{v_c^2}{\sqrt{\ln 2 - \frac{1}{2}}} \frac{\ln(1 + r/r_s)}{r/r_s}\\
        r^2 = x^2 + y^2 + z^2/q_z^2

    Parameters
    ----------
    v_c : numeric
        Circular velocity at the scale radius.
    r_s : numeric
        Scale radius.
    q_z : numeric
        Flattening.
    units : `~gala.units.UnitSystem` (optional)
        Set of non-reducable units that specify (at minimum) the
        length, mass, time, and angle units.

    """
    def __init__(self, v_c, r_s, q_z, units):
        parameters = OrderedDict()
        parameters['v_c'] = v_c
        parameters['r_s'] = r_s
        parameters['q_z'] = q_z
        super(FlattenedNFWPotential, self).__init__(parameters=parameters,
                                                    units=units)

# ============================================================================
#

cdef class LeeSutoTriaxialNFWWrapper(CPotentialWrapper):

    def __init__(self, G, parameters):
        cdef CPotential cp

        # This is the only code that needs to change per-potential
        cp.value[0] = <valuefunc>(leesuto_value)
        cp.density[0] = <densityfunc>(leesuto_density)
        cp.gradient[0] = <gradientfunc>(leesuto_gradient)
        # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

        cp.n_components = 1
        self._params = np.array([G] + list(parameters), dtype=np.float64)
        self._n_params = np.array([len(self._params)], dtype=np.int32)
        cp.n_params = &(self._n_params[0])
        cp.parameters[0] = &(self._params[0])
        cp.n_dim = 3
        self.cpotential = cp

class LeeSutoTriaxialNFWPotential(CPotentialBase):
    r"""
    LeeSutoTriaxialNFWPotential(v_c, r_s, a, b, c, units)

    Approximation of a Triaxial NFW Potential with the flattening in the density,
    not the potential. See Lee & Suto (2003) for details.

    See: http://adsabs.harvard.edu/abs/2003ApJ...585..151L

    Parameters
    ----------
    v_c : numeric
        Circular velocity.
    r_s : numeric
        Scale radius.
    a : numeric
        Major axis.
    b : numeric
        Intermediate axis.
    c : numeric
        Minor axis.
    units : `~gala.units.UnitSystem` (optional)
        Set of non-reducable units that specify (at minimum) the
        length, mass, time, and angle units.

    """
    def __init__(self, v_c, r_s, a, b, c, units):
        parameters = OrderedDict()
        parameters['v_c'] = v_c
        parameters['r_s'] = r_s
        parameters['a'] = a
        parameters['b'] = b
        parameters['c'] = c
        super(LeeSutoTriaxialNFWPotential, self).__init__(parameters=parameters,
                                                          units=units)

# ============================================================================

cdef class LogarithmicWrapper(CPotentialWrapper):

    def __init__(self, G, parameters):
        cdef CPotential cp

        # This is the only code that needs to change per-potential
        cp.value[0] = <valuefunc>(logarithmic_value)
        cp.density[0] = <densityfunc>(nan_density)
        cp.gradient[0] = <gradientfunc>(logarithmic_gradient)
        # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

        cp.n_components = 1
        self._params = np.array([G] + list(parameters), dtype=np.float64)
        self._n_params = np.array([len(self._params)], dtype=np.int32)
        cp.n_params = &(self._n_params[0])
        cp.parameters[0] = &(self._params[0])
        cp.n_dim = 3
        self.cpotential = cp

class LogarithmicPotential(CPotentialBase):
    r"""
    LogarithmicPotential(v_c, r_h, q1, q2, q3, phi=0, theta=0, psi=0, units)

    Triaxial logarithmic potential.

    .. math::

        \Phi(x,y,z) &= \frac{1}{2}v_{c}^2\ln((x/q_1)^2 + (y/q_2)^2 + (z/q_3)^2 + r_h^2)\\

    Parameters
    ----------
    v_c : `~astropy.units.Quantity`, numeric
        Circular velocity.
    r_h : `~astropy.units.Quantity`, numeric
        Scale radius.
    q1 : numeric
        Flattening in X.
    q2 : numeric
        Flattening in Y.
    q3 : numeric
        Flattening in Z.
    phi : `~astropy.units.Quantity`, numeric
        First euler angle in the z-x-z convention.
    units : `~gala.units.UnitSystem` (optional)
        Set of non-reducable units that specify (at minimum) the
        length, mass, time, and angle units.

    """
    def __init__(self, v_c, r_h, q1, q2, q3, phi=0., units=None):
        parameters = OrderedDict()
        parameters['v_c'] = v_c
        parameters['r_h'] = r_h
        parameters['q1'] = q1
        parameters['q2'] = q2
        parameters['q3'] = q3
        parameters['phi'] = phi
        super(LogarithmicPotential, self).__init__(parameters=parameters,
                                                   units=units)

        if not isinstance(self.units, DimensionlessUnitSystem):
            if self.units['angle'] != u.radian:
                raise ValueError("Angle unit must be radian.")

# ============================================================================
# TODO: why do these have to be in this file?

cdef class CCompositePotentialWrapper(CPotentialWrapper):

    def __init__(self, list potentials):
        cdef:
            CPotential cp
            CPotential tmp_cp
            int i
            CPotentialWrapper[::1] _cpotential_arr

        _cpotential_arr = np.array(potentials)

        n_components = len(potentials)
        self._n_params = np.zeros(n_components, dtype=np.int32)
        for i in range(n_components):
            self._n_params[i] = _cpotential_arr[i]._n_params[0]

        cp.n_components = n_components
        cp.n_params = &(self._n_params[0])
        cp.n_dim = 0

        for i in range(n_components):
            tmp_cp = _cpotential_arr[i].cpotential
            cp.parameters[i] = &(_cpotential_arr[i]._params[0])
            cp.value[i] = tmp_cp.value[0]
            cp.density[i] = tmp_cp.density[0]
            cp.gradient[i] = tmp_cp.gradient[0]

            if cp.n_dim == 0:
                cp.n_dim = tmp_cp.n_dim
            elif cp.n_dim != tmp_cp.n_dim:
                raise ValueError("Input potentials must have same number of coordinate dimensions")

        self.cpotential = cp

class CCompositePotential(CPotentialBase, CompositePotential):

    def __init__(self, **potentials):
        CompositePotential.__init__(self, **potentials)

    def _reset_c_instance(self):
        self._potential_list = []
        for p in self.values():
            self._potential_list.append(p.c_instance)
        self.G = p.G
        self.c_instance = CCompositePotentialWrapper(self._potential_list)

    def __setitem__(self, *args, **kwargs):
        CompositePotential.__setitem__(self, *args, **kwargs)
        self._reset_c_instance()

    def __reduce__(self):
        """ Properly package the object for pickling """
        derp = tuple([self.units] + [c.parameters for c in self.values()])
        return (self.__class__, derp)



# coding: utf-8

from __future__ import division, print_function

__author__ = "adrn <adrn@astro.columbia.edu>"

# Third-party
import numpy as np
import pytest

# Project
from ... import potential as gp
from ..nonlinear import lyapunov_max, fast_lyapunov_max, surface_of_section
from ...integrate import DOPRI853Integrator
from ...util import atleast_2d
from ...units import galactic

class TestForcedPendulum(object):

    def setup(self):

        def F(t,x,A,omega_d):
            q,p = x
            return np.array([p,-np.sin(q) + A*np.cos(omega_d*t)])

        # initial conditions and parameter choices for chaotic / regular pendulum
        self.regular_w0 = np.array([1.,0.])
        self.regular_par = (0.055, 0.7)
        self.regular_integrator = DOPRI853Integrator(F, func_args=self.regular_par)

        self.chaotic_w0 = np.array([3.,0.])
        self.chaotic_par = (0.07, 0.75)
        self.chaotic_integrator = DOPRI853Integrator(F, func_args=self.chaotic_par)

    def test_lyapunov_max(self, tmpdir):
        n_steps = 20000
        dt = 1.
        n_steps_per_pullback = 10
        d0 = 1e-5
        noffset = 2

        regular_LEs, regular_orbit = lyapunov_max(self.regular_w0, self.regular_integrator,
                                                  dt=dt, n_steps=n_steps,
                                                  d0=d0, n_steps_per_pullback=n_steps_per_pullback,
                                                  noffset_orbits=noffset)

        regular_LEs = np.mean(regular_LEs, axis=1)
        assert regular_LEs[-1] < 1E-3

        chaotic_LEs, chaotic_orbit = lyapunov_max(self.chaotic_w0, self.chaotic_integrator,
                                                  dt=dt, n_steps=n_steps,
                                                  d0=d0, n_steps_per_pullback=n_steps_per_pullback,
                                                  noffset_orbits=noffset)
        chaotic_LEs = np.mean(chaotic_LEs, axis=1)
        assert chaotic_LEs[-1] > 1E-2

        # pl.figure()
        # pl.loglog(regular_LEs, marker=None)
        # pl.savefig(os.path.join(str(tmpdir),"pend_regular.png"))

        # pl.figure()
        # pl.plot(t, regular_ws[:,0], marker=None)
        # pl.savefig(os.path.join(str(tmpdir),"pend_orbit_regular.png"))

        # pl.figure()
        # pl.loglog(chaotic_LEs, marker=None)
        # pl.savefig(os.path.join(str(tmpdir),"pend_chaotic.png"))

        # pl.figure()
        # pl.plot(t, chaotic_ws[:,0], marker=None)
        # pl.savefig(os.path.join(str(tmpdir),"pend_orbit_chaotic.png"))

        # pl.close('all')

# --------------------------------------------------------------------

class HenonHeilesBase(object):

    def potential(self,w,A,B,C,D):
        x,y = w[:2]
        term1 = 0.5*(A*x**2 + B*y**2)
        term2 = D*x**2*y - C/3.*y**3
        return term1 + term2

    def acceleration(self,w,A,B,C,D):
        x,y = w[:2]
        ax = -(A*x + 2*D*x*y)
        ay = -(B*y + D*x*x - C*y*y)
        return np.array([ax, ay])

    def jerk(self,w,A,B,C,D):
        x,y = w[:2]
        dx,dy = w[4:6]

        dax = -(A+2*D*y)*dx - 2*D*x*dy
        day = -2*D*x*dx - (B-2*C*y)*dy

        return np.array([dax,day])

    def F_max(self,t,w,*args):
        x,y,px,py = w
        term1 = np.array([px, py])
        term2 = self.acceleration(w, *args)
        return np.vstack((term1,term2))

    def setup(self):
        # parameter choices
        self.par = (1.,1.,1.,1.)
        self.n_steps = 2000
        self.dt = 2.

    def test_integrate_orbit(self, tmpdir):
        integrator = DOPRI853Integrator(self.F_max, func_args=self.par)

        # pl.clf()
        orbit = integrator.run(self.w0, dt=self.dt, n_steps=self.n_steps)
        # pl.plot(w[:,0,0], w[:,0,1], marker=None)

        # pl.savefig(os.path.join(str(tmpdir),"hh_orbit_{}.png".format(self.__class__.__name__)))

    def test_lyapunov_max(self, tmpdir):
        n_steps_per_pullback = 10
        d0 = 1e-5
        noffset = 2

        integrator = DOPRI853Integrator(self.F_max, func_args=self.par)
        lyap, orbit = lyapunov_max(self.w0, integrator,
                                   dt=self.dt, n_steps=self.n_steps,
                                   d0=d0, noffset_orbits=noffset,
                                   n_steps_per_pullback=n_steps_per_pullback)
        lyap = np.mean(lyap, axis=1)

        # pl.clf()
        # pl.loglog(lyap, marker=None)
        # pl.savefig(os.path.join(str(tmpdir),"hh_lyap_max_{}.png".format(self.__class__.__name__)))

        # pl.clf()
        # pl.plot(ws[...,0], ws[...,1], marker=None)
        # pl.savefig(os.path.join(str(tmpdir),"hh_orbit_lyap_max_{}.png".format(self.__class__.__name__)))

# initial conditions from LP-VI documentation:
class TestHenonHeilesStablePeriodic(HenonHeilesBase):
    def setup(self):
        super(TestHenonHeilesStablePeriodic, self).setup()
        self.w0 = np.array([0.,0.295456,0.407308431,0.])
        self.check = lambda x: x < 1E-3

class TestHenonHeilesStableQuasi1(HenonHeilesBase):
    def setup(self):
        super(TestHenonHeilesStableQuasi1, self).setup()
        self.w0 = np.array([0., 0.483, 0.27898039, 0.])
        self.check = lambda x: x < 2E-3

class TestHenonHeilesStableQuasi2(HenonHeilesBase):
    def setup(self):
        super(TestHenonHeilesStableQuasi2, self).setup()
        self.w0 = np.array([0., 0.46912, 0.291124891, 0.])
        self.check = lambda x: x < 2E-3

class TestHenonHeilesStableChaos1(HenonHeilesBase):
    def setup(self):
        super(TestHenonHeilesStableChaos1, self).setup()
        self.w0 = np.array([0., 0.509, 0.254624859, 0.])
        self.check = lambda x: x > 2E-3

class TestHenonHeilesStableChaos2(HenonHeilesBase):
    def setup(self):
        super(TestHenonHeilesStableChaos2, self).setup()
        self.w0 = np.array([0., 0.56, 0.164113781, 0.112])
        self.check = lambda x: x > 1E-2

# --------------------------------------------------------------------

class TestLogarithmic(object):

    def F(self,t,w):
        x,y,z,px,py,pz = w
        term1 = atleast_2d([px, py, pz], insert_axis=1)
        term2 = self.potential.acceleration(w[:3])
        return np.vstack((term1,term2))

    def setup(self):

        # set the potential
        self.potential = gp.LogarithmicPotential(v_c=np.sqrt(2), r_h=0.1,
                                                 q1=1., q2=0.9, q3=1.,
                                                 units=galactic)

        # see figure 1 from Papaphillipou & Laskar
        x0 = -0.01
        X0 = -0.2
        y0 = 0.
        E0 = -0.4059
        Y0 = np.sqrt(E0 - self.potential.value([x0,y0,0.]).value)
        chaotic_w0 = [x0,y0,0.,X0,Y0,0.]

        # initial conditions from LP-VI documentation:
        self.w0s = np.array([[0.49, 0., 0., 1.3156, 0.4788, 0.],  # regular
                             chaotic_w0])  # chaotic

        self.n_steps = 25000
        self.dt = 0.004

    def test_fast_lyapunov_max(self, tmpdir):
        n_steps_per_pullback = 10
        d0 = 1e-5
        noffset = 2

        for ii,w0 in enumerate(self.w0s):
            print(ii, w0)
            lyap, orbit = fast_lyapunov_max(w0, self.potential,
                                            dt=self.dt, n_steps=self.n_steps,
                                            d0=d0, noffset_orbits=noffset,
                                            n_steps_per_pullback=n_steps_per_pullback)
            lyap = np.mean(lyap, axis=1)

            # also just integrate the orbit to compare dE scaling
            orbit2 = self.potential.integrate_orbit(w0, dt=self.dt, n_steps=self.n_steps,
                                                    Integrator=DOPRI853Integrator)

            # lyapunov exp
            # pl.figure()
            # pl.loglog(lyap, marker=None)
            # pl.savefig(os.path.join(str(tmpdir),"log_lyap_max_{}.png".format(ii)))

            # energy conservation
            E = orbit[:,0].energy().value # returns 3 orbits
            dE = np.abs(E[1:] - E[0])

            E = orbit2.energy().value
            dE_ww = np.abs(E[1:] - E[0])

            assert np.allclose(dE_ww[-100:], dE[-100:], rtol=1E-1)

            # pl.figure()
            # pl.semilogy(dE_ww, marker=None)
            # pl.semilogy(dE, marker=None)
            # pl.show()
            # pl.savefig(os.path.join(str(tmpdir),"log_dE_{}.png".format(ii)))

            # pl.figure(figsize=(6,6))
            # pl.plot(ws[:,0], ws[:,1], marker='.', linestyle='none', alpha=0.1)
            # pl.savefig(os.path.join(str(tmpdir),"log_orbit_lyap_max_{}.png".format(ii)))

    @pytest.mark.slow
    def test_compare_fast(self, tmpdir):
        n_steps_per_pullback = 10
        d0 = 1e-5
        noffset = 2

        integrator = DOPRI853Integrator(self.F)
        for ii,w0 in enumerate(self.w0s):
            print(ii)

            lyap1, orbit1 = fast_lyapunov_max(w0, self.potential,
                                              dt=self.dt, n_steps=self.n_steps,
                                              d0=d0, noffset_orbits=noffset,
                                              n_steps_per_pullback=n_steps_per_pullback)
            lyap1 = np.mean(lyap1, axis=1)

            # check energy conservation
            E = orbit1.energy().value
            dE_fast = np.abs(E[1:] - E[0])
            assert np.all(dE_fast[:,0] < 1E-10)

            lyap2, orbit2 = lyapunov_max(w0.copy(), integrator,
                                         dt=self.dt, n_steps=self.n_steps//10,
                                         d0=d0, noffset_orbits=noffset,
                                         n_steps_per_pullback=n_steps_per_pullback,
                                         units=self.potential.units)
            lyap2 = np.mean(lyap2, axis=1)

            # check energy conservation
            E = orbit2.energy(self.potential).value
            dE_slow = np.abs(E[1:] - E[0])
            assert np.all(dE_slow[:,0] < 1E-10)

            # lyapunov exp
            # pl.clf()
            # pl.loglog(t1[1:-10:10], lyap1, marker=None)
            # pl.loglog(t2[1:-10:10], lyap2, marker=None)
            # pl.show()
            # pl.savefig(os.path.join(str(tmpdir),"log_lyap_compare_{}.png".format(ii)))

            # pl.clf()
            # pl.semilogy(dE_ww, marker=None)
            # pl.semilogy(dE, marker=None)
            # pl.savefig(os.path.join(str(tmpdir),"log_dE_{}.png".format(ii)))

            # pl.figure(figsize=(6,6))
            # pl.plot(ws[:,0], ws[:,1], marker='.', linestyle='none', alpha=0.1)
            # pl.savefig(os.path.join(str(tmpdir),"log_orbit_lyap_max_{}.png".format(ii)))

@pytest.mark.skipif(True, reason="too slow")
def test_surface_of_section(tmpdir):
    # TODO: needs overhaul
    # from mpl_toolkits.mplot3d import Axes3D
    from ...potential import LogarithmicPotential
    from ...units import galactic

    pot = LogarithmicPotential(v_c=1., r_h=1., q1=1., q2=0.9, q3=0.8, units=galactic)

    w0 = np.array([[0.,0.8,0.,1.,0.,0.],
                   [0.,0.9,0.,1.,0.,0.]]).T
    orbit = pot.integrate_orbit(w0, dt=0.02, n_steps=100000)
    sos = surface_of_section(orbit, plane_ix=1)

    # plot in 3D
    # fig = pl.figure(figsize=(10,10))
    # ax = fig.add_subplot(111, projection='3d')

    # ax.scatter(np.concatenate(sos[0]), # x
    #            np.concatenate(sos[3]), # xdot
    #            np.concatenate(sos[2]), # z
    #            c=np.concatenate(sos[5])) # zdot
    # ax.set_xlabel('$x$')
    # ax.set_ylabel(r'$\dot{x}$')
    # ax.set_zlabel('$z$')

    # fig.tight_layout()
    # pl.show()
    # fig.savefig(os.path.join("sos.png"))



# coding: utf-8

""" Test velocity transformations.  """

from __future__ import division, print_function

__author__ = "adrn <adrn@astro.columbia.edu>"

# Standard library
import logging

# Third-party
import numpy as np
from astropy import log as logger
import astropy.coordinates as coord
import astropy.units as u

# This project
from ..velocity_transforms import *

logger.setLevel(logging.DEBUG)

class TestTransforms(object):

    def setup(self):
        self.pos = ([[10.,0,0], [10,0,0], [4,4,0]] * u.kpc).T
        self.vel = ([[0.,100,500], [0,-100,-500], [100.,100.,-500]] * u.km/u.s).T

        self.pos_repr = coord.CartesianRepresentation(self.pos)

    def test_cartesian_to(self):
        for i,func in enumerate([cartesian_to_spherical,
                                 cartesian_to_physicsspherical,
                                 cartesian_to_cylindrical]):

            # dimensionless
            vsph1 = func(self.pos.value * u.dimensionless_unscaled,
                         self.vel.value * u.dimensionless_unscaled)
            assert vsph1.unit == u.dimensionless_unscaled

            # astropy coordinates
            cpos = coord.SkyCoord(self.pos_repr)
            vsph2 = func(cpos, self.vel)
            assert vsph2.unit == u.km/u.s

            # astropy representation
            vsph3 = func(self.pos_repr, self.vel)
            assert vsph3.unit == u.km/u.s

            np.testing.assert_allclose(vsph1.value, vsph2.value, atol=1E-10)
            np.testing.assert_allclose(vsph1.value, vsph3.value, atol=1E-10)

            true_v = self.vel.copy()
            if func == cartesian_to_physicsspherical:
                true_v[2] *= -1.

            np.testing.assert_allclose(vsph1[:,:2].value, true_v[:,:2].value)

            assert vsph1[0,2] > 0.  # vr
            assert np.sign(vsph1[2,2]) == np.sign(true_v[2,2])
            np.testing.assert_allclose(np.sqrt(np.sum(vsph1**2,axis=0)).value,
                                       np.sqrt(np.sum(true_v**2,axis=0)).value)

    def test_to_cartesian(self):
        for i,func in enumerate([spherical_to_cartesian,
                                 physicsspherical_to_cartesian,
                                 cylindrical_to_cartesian]):
            true_v = ([[0.,100,500], [0,-100,-500], [0.,np.sqrt(2)*100,-500]] * u.km/u.s).T

            # dimensionless
            vsph1 = func(self.pos.value * u.dimensionless_unscaled,
                         self.vel.value * u.dimensionless_unscaled)
            assert vsph1.unit == u.dimensionless_unscaled

            # astropy coordinates
            cpos = coord.SkyCoord(self.pos_repr)
            vsph2 = func(cpos, self.vel)
            assert vsph2.unit == u.km/u.s

            # astropy representation
            vsph3 = func(self.pos_repr, self.vel)
            assert vsph3.unit == u.km/u.s

            np.testing.assert_allclose(vsph1.value, vsph2.value, atol=1E-10)
            np.testing.assert_allclose(vsph1.value, vsph3.value, atol=1E-10)

            if func == physicsspherical_to_cartesian:
                true_v[2] *= -1.

            np.testing.assert_allclose(vsph1.value, true_v.value, atol=1E-10)



cdef extern from "src/cpotential.h":
    ctypedef struct CPotential:
        pass

cdef class CPotentialWrapper:
    cdef CPotential cpotential
    cdef double[::1] _params
    cdef int[::1] _n_params

    cpdef value(self, double[:,::1] q, double t=?)
    cpdef density(self, double[:,::1] q, double t=?)
    cpdef gradient(self, double[:,::1] q, double t=?)
    # cpdef hessian(self, double[:,::1] q, double t=?)

    cpdef d_dr(self, double[:,::1] q, double G, double t=?)
    cpdef d2_dr2(self, double[:,::1] q, double G, double t=?)
    cpdef mass_enclosed(self, double[:,::1] q, double G, double t=?)



from .core import PhaseSpacePosition, CartesianPhaseSpacePosition
from .orbit import Orbit, CartesianOrbit
from .analyticactionangle import *
from .actionangle import *
from .nonlinear import *
from .plot import *
from .util import *



# coding: utf-8

""" ...explain... """

from __future__ import division, print_function

__author__ = "adrn <adrn@astro.columbia.edu>"

# Third-party
import numpy as np

# Project
from ..util import atleast_2d

__all__ = ['plot_orbits', 'three_panel']

def _get_axes(dim, axes=None, triangle=False, subplots_kwargs=dict()):
    """
    Parameters
    ----------
    dim : int
        Dimensionality of the orbit.
    axes : array_like (optional)
        Array of matplotlib Axes objects.
    triangle : bool (optional)
        Make a triangle plot instead of plotting all projections in a single row.
    subplots_kwargs : dict (optional)
        Dictionary of kwargs passed to :func:`~matplotlib.pyplot.subplots`.
    """

    import matplotlib.pyplot as plt

    if dim == 3:
        if triangle and axes is None:
            figsize = subplots_kwargs.pop('figsize', (8,8))
            sharex = subplots_kwargs.pop('sharex', True)
            sharey = subplots_kwargs.pop('sharey', True)
            fig,axes = plt.subplots(2,2,figsize=figsize, sharex=sharex, sharey=sharey,
                                    **subplots_kwargs)
            axes[0,1].set_visible(False)
            axes = axes.flat
            axes = [axes[0],axes[2],axes[3]]

        elif triangle and axes is not None:
            try:
                axes = axes.flat
            except:
                pass

            if len(axes) == 4:
                axes = [axes[0],axes[2],axes[3]]

        elif not triangle and axes is None:
            figsize = subplots_kwargs.pop('figsize', (10,3.5))
            fig,axes = plt.subplots(1, 3, figsize=figsize, **subplots_kwargs)

    elif dim <= 2:
        if axes is not None:
            try:
                len(axes)
            except TypeError:  # single axes object
                axes = [axes]

        else:
            if dim ==1:
                figsize = subplots_kwargs.pop('figsize', (14,8))
            elif dim == 2:
                figsize = subplots_kwargs.pop('figsize', (8,8))

            fig,axes = plt.subplots(1, 1, figsize=figsize, **subplots_kwargs)
            axes = [axes]

    else:
        raise ValueError("Orbit must have dimensions <= 3.")

    return axes

def plot_orbits(x, t=None, ix=None, axes=None, triangle=False,
                subplots_kwargs=dict(), labels=("$x$", "$y$", "$z$"), **kwargs):
    """
    Given time series of positions, `x`, make nice plots of the orbit in
    cartesian projections.

    Parameters
    ----------
    x : array_like
        Array of positions. ``axis=0`` is assumed to be the dimensionality,
        ``axis=1`` is the time axis. See :ref:`shape-conventions` for more information.
    t : array_like (optional)
        Array of times. Only used if the input orbit is 1-dimensional.
    ix : int, array_like (optional)
        Index or array of indices of orbits to plot. For example, if `x` is an
        array of shape ``(3,1024,32)`` - 1024 timesteps for 32 orbits in 3D
        positions -- `ix` would specify which of the 32 orbits to plot.
    axes : array_like (optional)
        Array of matplotlib Axes objects.
    triangle : bool (optional)
        Make a triangle plot instead of plotting all projections in a single row.
    subplots_kwargs : dict (optional)
        Dictionary of kwargs passed to :func:`~matplotlib.pyplot.subplots`.
    labels : iterable (optional)
        List or iterable of axis labels as strings. They should correspond to the
        dimensions of the input orbit.
    **kwargs
        All other keyword arguments are passed to :func:`~matplotlib.pyplot.plot`.
        You can pass in any of the usual style kwargs like ``color=...``,
        ``marker=...``, etc.

    Returns
    -------
    fig : `~matplotlib.Figure`
    """

    x = atleast_2d(x, insert_axis=1)
    if x.ndim == 2:
        x = x[...,np.newaxis]

    # dimensionality of input orbit
    dim = x.shape[0]
    if dim > 3:
        # if orbit has more than 3 dimensions, only use the first 3
        dim = 3

    # hack in some defaults to subplots kwargs so by default share x and y axes
    if 'sharex' not in subplots_kwargs:
        subplots_kwargs['sharex'] = True

    if 'sharey' not in subplots_kwargs:
        subplots_kwargs['sharey'] = True

    axes = _get_axes(dim=dim, axes=axes, triangle=triangle,
                     subplots_kwargs=subplots_kwargs)

    if ix is not None:
        ixs = np.atleast_1d(ix)
    else:
        ixs = range(x.shape[-1])

    if dim == 3:
        for ii in ixs:
            axes[0].plot(x[0,:,ii], x[1,:,ii], **kwargs)
            axes[1].plot(x[0,:,ii], x[2,:,ii], **kwargs)
            axes[2].plot(x[1,:,ii], x[2,:,ii], **kwargs)

        if triangle:
            # HACK: until matplotlib 1.4 comes out, need this
            axes[0].set_ylim(axes[0].get_xlim())
            axes[2].set_xlim(axes[0].get_ylim())

            axes[0].set_ylabel(labels[1])
            axes[1].set_xlabel(labels[0])
            axes[1].set_ylabel(labels[2])
            axes[2].set_xlabel(labels[1])

        else:
            axes[0].set_xlabel(labels[0])
            axes[0].set_ylabel(labels[1])

            axes[1].set_xlabel(labels[0])
            axes[1].set_ylabel(labels[2])

            axes[2].set_xlabel(labels[1])
            axes[2].set_ylabel(labels[2])

        if not triangle:
            axes[0].figure.tight_layout()

    elif dim == 2:
        for ii in ixs:
            axes[0].plot(x[0,:,ii], x[1,:,ii], **kwargs)

        axes[0].set_xlabel(labels[0])
        axes[0].set_ylabel(labels[1])
        axes[0].figure.tight_layout()

    elif dim == 1:
        if t is None:
            t = np.arange(x.shape[1])

        for ii in ixs:
            axes[0].plot(t, x[0,:,ii], **kwargs)

        axes[0].set_xlabel("$t$")
        axes[0].set_ylabel(labels[0])
        axes[0].figure.tight_layout()

    return axes[0].figure

def three_panel(q, relative_to=None, autolim=True, axes=None,
                triangle=False, subplots_kwargs=dict(), labels=None, **kwargs):
    """
    Given 3D quantities, ``q``, make a nice three-panel or triangle plot
    of projections of the values.

    Parameters
    ----------
    q : array_like
        Array of values. ``axis=0`` is assumed to be the dimensionality,
        ``axis=1`` is the time axis. See :ref:`shape-conventions` for more information.
    relative_to : bool (optional)
        Plot the values relative to this value or values.
    autolim : bool (optional)
        Automatically set the plot limits to be something sensible.
    axes : array_like (optional)
        Array of matplotlib Axes objects.
    triangle : bool (optional)
        Make a triangle plot instead of plotting all projections in a single row.
    subplots_kwargs : dict (optional)
        Dictionary of kwargs passed to :func:`~matplotlib.pyplot.subplots`.
    labels : iterable (optional)
        List or iterable of axis labels as strings. They should correspond to the
        dimensions of the input orbit.
    **kwargs
        All other keyword arguments are passed to :func:`~matplotlib.pyplot.scatter`.
        You can pass in any of the usual style kwargs like ``color=...``,
        ``marker=...``, etc.

    Returns
    -------
    fig : `~matplotlib.Figure`
    """

    # don't propagate changes back...
    q = q.copy()

    # get axes object from arguments
    axes = _get_axes(dim=3, axes=axes, triangle=triangle, subplots_kwargs=subplots_kwargs)

    # if the quantities are relative
    if relative_to is not None:
        q -= relative_to

    axes[0].scatter(q[0], q[1], **kwargs)
    axes[1].scatter(q[0], q[2], **kwargs)
    axes[2].scatter(q[1], q[2], **kwargs)

    if labels is not None:
        if triangle:
            axes[0].set_ylabel(labels[1])
            axes[1].set_xlabel(labels[0])
            axes[1].set_ylabel(labels[2])
            axes[2].set_xlabel(labels[0])

        else:
            axes[0].set_xlabel(labels[0])
            axes[0].set_ylabel(labels[1])

            axes[1].set_xlabel(labels[0])
            axes[1].set_ylabel(labels[2])

            axes[2].set_xlabel(labels[1])
            axes[2].set_ylabel(labels[2])

    if autolim:
        lims = []
        for i in range(3):
            mx,mi = q[i].max(), q[i].min()
            delta = mx-mi
            lims.append((mi-delta*0.05, mx+delta*0.05))

        axes[0].set_xlim(lims[0])
        axes[0].set_ylim(lims[1])
        axes[1].set_xlim(lims[0])
        axes[1].set_ylim(lims[2])
        axes[2].set_xlim(lims[1])
        axes[2].set_ylim(lims[2])

    if not triangle:
        axes[0].figure.tight_layout()
    return axes[0].figure



# coding: utf-8

""" Astropy coordinate class for the Sagittarius coordinate system """

from __future__ import division, print_function

__author__ = "adrn <adrn@astro.columbia.edu>"

# Third-party
import numpy as np

from astropy.coordinates import frame_transform_graph
from astropy.coordinates.angles import rotation_matrix
import astropy.coordinates as coord
import astropy.units as u

__all__ = ["Orphan"]

class Orphan(coord.BaseCoordinateFrame):
    """
    A Heliocentric spherical coordinate system defined by the orbit
    of the Orphan stream, as described in
    Newberg et al. 2010 (see: `<http://arxiv.org/abs/1001.0576>`_).

    For more information about this class, see the Astropy documentation
    on coordinate frames in :mod:`~astropy.coordinates`.

    Parameters
    ----------
    representation : :class:`~astropy.coordinates.BaseRepresentation` or None
        A representation object or None to have no data (or use the other keywords)
    Lambda : angle_like, optional, must be keyword
        The longitude-like angle corresponding to Orphan's orbit.
    Beta : angle_like, optional, must be keyword
        The latitude-like angle corresponding to Orphan's orbit.
    distance : :class:`~astropy.units.Quantity`, optional, must be keyword
        The Distance for this object along the line-of-sight.

    """
    default_representation = coord.SphericalRepresentation

    frame_specific_representation_info = {
        'spherical': [coord.RepresentationMapping('lon', 'Lambda'),
                      coord.RepresentationMapping('lat', 'Beta'),
                      coord.RepresentationMapping('distance', 'distance')],
        'unitspherical': [coord.RepresentationMapping('lon', 'Lambda'),
                          coord.RepresentationMapping('lat', 'Beta')]
    }

# Define the Euler angles
phi = np.radians(128.79)
theta = np.radians(54.39)
psi = np.radians(90.70)

# Generate the rotation matrix using the x-convention (see Goldstein)
D = rotation_matrix(phi, "z", unit=u.radian)
C = rotation_matrix(theta, "x", unit=u.radian)
B = rotation_matrix(psi, "z", unit=u.radian)
R = np.array(B.dot(C).dot(D))

@frame_transform_graph.transform(coord.StaticMatrixTransform, coord.Galactic, Orphan)
def galactic_to_orp():
    """ Compute the transformation from Galactic spherical to
        heliocentric Orphan coordinates.
    """
    return R

# Oph to Galactic coordinates
@frame_transform_graph.transform(coord.StaticMatrixTransform, Orphan, coord.Galactic)
def oph_to_galactic():
    """ Compute the transformation from heliocentric Orphan coordinates to
        spherical Galactic.
    """
    return galactic_to_orp().T
